from botocore import exceptions as botoerror

from .base_control import (
    DeleteResourceDialog,
    Describer,
    ListResourceDocumentCreator,
    ResourceLister,
    SingleNameDialog,
    SingleSelectorDialog,
)
from .common import Common
from .termui.alignment import CenterAnchor, Dimension
from .termui.common import Commons
from .termui.ui import ControlCodes


class InlinePolicyAttacher:
    def __init__(self, parent, datatype):
        self.datatype = datatype
        self.parent = parent
        self.client = Common.Session.service_provider("iam")

        self.get = getattr(self.client, "get_{0}_policy".format(datatype))
        self.put_method = "put_{0}_policy".format(datatype)
        self.arg = "{0}Name".format(datatype.capitalize())

    def attach_inline_policy(self, _):
        if self.parent.selection is None:
            return
        SingleNameDialog(
            self.parent.parent,
            "Attach new or edit existing inline policy for {0}".format(self.datatype),
            self.do_attach_inline,
            label="Policy name:",
            what="name",
            subject="policy",
            default="",
            caller=self.parent,
            accepted_inputs=Commons.alphanum_with_symbols("+=,.@_-"),
        )

    def do_attach_inline(self, name, selection_is_policy=False):
        if self.parent.selection is None:
            return
        if selection_is_policy:
            kwargs = {"PolicyName": self.parent.selection["name"], self.arg: name}
        else:
            kwargs = {self.arg: self.parent.selection["name"], "PolicyName": name}
        try:
            chk = self.get(**kwargs)
            initial_document = {"PolicyDocument": chk["PolicyDocument"]}
        except self.client.exceptions.NoSuchEntityException:
            initial_document = {
                "PolicyDocument": {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Deny",
                            "Action": [""],
                            "Resource": "*",
                            "Condition": {},
                        }
                    ],
                }
            }
        except botoerror.ClientError as e:
            Common.clienterror(
                e,
                "Retrieve Inline {0} Policy".format(self.datatype.capitalize()),
                "IAM",
                set_message=True,
            )
            return
        except Exception as e:
            Common.error(
                str(e),
                "Retrieve Inline {0} Policy".format(self.datatype.capitalize()),
                "IAM",
                set_message=True,
            )
            return

        creator = ListResourceDocumentCreator(
            "iam",
            self.put_method,
            None,
            initial_document,
            as_json=False,
            static_fields=kwargs,
            message="Put policy successful",
        )
        creator.edit()
        self.parent.refresh_data()


class UserLister(ResourceLister):
    prefix = "user_list"
    title = "Users"
    command_palette = ["user", "users"]

    def title_info(self):
        return (
            "Group: {0}".format(self.group["name"]) if self.group is not None else None
        )

    def create_user(self, _):
        creator = ListResourceDocumentCreator(
            "iam",
            "create_user",
            None,
            as_json=["Tags"],
            initial_document={
                "UserName": "",
                "Path": "/",
                "PermissionsBoundary": "",
                "Tags": [{"Key": "", "Value": ""}],
            },
        )
        creator.edit()
        self.refresh_data()

    def delete_user(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="user",
            resource_identifier=self.selection["name"],
            callback=self.do_delete,
            action_name="Delete",
        )

    def do_delete(self, **kwargs):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "delete_user",
            {
                "UserName": self.selection["name"],
            },
            "Delete User",
            "IAM",
            subcategory="User",
            success_template="Deleting user {UserName}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def do_attach(self, policy_arn):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "attach_user_policy",
            {
                "UserName": self.selection["name"],
                "PolicyArn": policy_arn,
            },
            "Attach Managed User Policy",
            "IAM",
            subcategory="User",
            success_template="Attaching policy {PolicyArn} to user {UserName}",
            resource=self.selection["name"],
        )

    def attach_policy(self, _):
        if self.selection is None:
            return
        SingleSelectorDialog(
            self.parent,
            "Attach policy to user",
            "policy",
            "attach",
            PolicyLister,
            self.do_attach,
            caller=self,
        )

    def add_user_to_group(self, _):
        if self.selection is None:
            return
        SingleSelectorDialog(
            self.parent,
            "Add user to group",
            "group",
            "add",
            GroupLister,
            self.do_add_to_group,
            caller=self,
        )

    def do_add_to_group(self, group):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "add_user_to_group",
            {"UserName": self.selection["name"], "GroupName": group},
            "Add User to Group",
            "IAM",
            subcategory="User",
            success_template="Adding user {UserName} to group {GroupName}",
            resource=self.selection["name"],
        )

    def remove_from_group(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="user",
            resource_identifier=self.selection["name"],
            from_what="group",
            from_what_name=self.group["name"],
            callback=self.do_remove,
            action_name="Remove",
        )

    def do_remove(self, **kwargs):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "remove_user_from_group",
            {"UserName": self.selection["name"], "GroupName": self.group["name"]},
            "Remove User from Group",
            "IAM",
            subcategory="User",
            success_template="Removing user {UserName} from group {GroupName}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def determine_login_profile(self, user):
        resp = Common.generic_api_call(
            "iam",
            "get_login_profile",
            {"UserName": user},
            "Get Login Profile",
            "IAM",
            subcategory="User",
            resource=user,
        )
        if resp["Success"]:
            return True
        return False

    def get_or_create_login_profile(self, _):
        if self.selection is None:
            return
        has_login_profile = self.determine_login_profile(self.selection["name"])
        if has_login_profile:
            Common.Session.push_frame(
                LoginProfileDescriber.opener(
                    entry=self.selection, caller=self, pushed=True
                )
            )
        else:
            SingleNameDialog(
                self.parent,
                "Create login profile for {0}".format(self.selection["name"]),
                self.do_create_login_profile,
                label="Password:",
                what="password",
                subject="login profile",
                default="",
                caller=self,
            )

    def do_create_login_profile(self, password):
        if self.selection is None:
            return
        resp = Common.generic_api_call(
            "iam",
            "create_login_profile",
            {
                "UserName": self.selection["name"],
                "Password": password,
                "PasswordResetRequired": True,
            },
            "Create Login Profile",
            "IAM",
            subcategory="User",
            resource=self.selection["name"],
        )
        if resp["Success"]:
            Common.Session.push_frame(
                LoginProfileDescriber.opener(
                    entry=self.selection, caller=self, pushed=True
                )
            )

    def __init__(self, *args, group=None, **kwargs):
        self.resource_key = "iam"
        self.group = group
        self.list_method = "list_users" if group is None else "get_group"
        self.list_kwargs = {"GroupName": group["name"]} if group is not None else {}
        self.item_path = ".Users"
        self.column_paths = {
            "name": ".UserName",
            "path": ".Path",
            "last login": ".PasswordLastUsed",
        }
        self.imported_column_sizes = {
            "name": 25,
            "path": 40,
            "last login": 15,
        }
        self.hidden_columns = {
            "arn": ".Arn",
        }
        self.describe_command = UserDescriber.opener
        self.open_command = "m"
        self.open_selection_arg = "user"
        self.additional_commands = {
            "m": {
                "command": PolicyLister.opener,
                "selection_arg": "user",
                "tooltip": "View Managed Policies",
            },
            "g": {
                "command": GroupLister.opener,
                "selection_arg": "user",
                "tooltip": "List Groups for User",
            },
            "p": {
                "command": InlinePolicyLister.opener,
                "selection_arg": "data",
                "tooltip": "View Inline Policies",
                "kwargs": {
                    "datatype": "user",
                },
            },
        }

        self.imported_column_order = ["name", "path", "last login"]
        self.sort_column = "name"
        super().__init__(*args, **kwargs)
        self.inliner = InlinePolicyAttacher(self, "user")
        self.add_hotkey("a", self.attach_policy, "Attach Policy")
        self.add_hotkey("c", self.create_user, "Create User")
        self.add_hotkey("i", self.inliner.attach_inline_policy, "Put Inline Policy")
        self.add_hotkey(
            "p", self.get_or_create_login_profile, "Get/Create Login Profile"
        )
        if self.group is None:
            self.add_hotkey("j", self.add_user_to_group, "Add to Group")
            self.add_hotkey(ControlCodes.D, self.delete_user, "Delete User")
        else:
            self.add_hotkey(ControlCodes.D, self.remove_from_group, "Remove from Group")


class UserDescriber(Describer):
    prefix = "user_browser"
    title = "User"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "iam"
        self.describe_method = "get_user"
        self.describe_kwarg_name = "UserName"
        self.object_path = ".User"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class LoginProfileDescriber(Describer):
    prefix = "login_profile_browser"
    title = "Login Profile"

    def title_info(self):
        return "User: {0}".format(self.user["name"])

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.user = entry
        self.resource_key = "iam"
        self.describe_method = "get_login_profile"
        self.describe_kwarg_name = "UserName"
        self.object_path = ".LoginProfile"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )
        self.add_hotkey(
            ControlCodes.D, self.delete_login_profile, "Delete Login Profile"
        )

    def delete_login_profile(self, _):
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="login profile",
            resource_identifier=self.user["name"],
            from_what="user",
            from_what_name=self.user["name"],
            callback=self.do_delete_login_profile,
            action_name="Delete",
        )

    def do_delete_login_profile(self, *args, **kwargs):
        resp = Common.generic_api_call(
            "iam",
            "delete_login_profile",
            {
                "UserName": self.user["name"],
            },
            "Delete Login Profile",
            "IAM",
            subcategory="User",
            resource=self.user["name"],
            success_template="Deleting login profile for {UserName}",
        )
        if resp["Success"]:
            Common.Session.pop_frame()


class GroupLister(ResourceLister):
    prefix = "group_list"
    title = "Groups"
    command_palette = ["group", "groups"]

    def title_info(self):
        return "User: {0}".format(self.user["name"]) if self.user is not None else None

    def do_attach(self, policy_arn):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "attach_group_policy",
            {"GroupName": self.selection["name"], "PolicyArn": policy_arn},
            "Attach Managed Group Policy",
            "IAM",
            subcategory="Group",
            success_template="Attaching policy {PolicyArn} to group {GroupName}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def attach_policy(self, _):
        if self.selection is None:
            return
        SingleSelectorDialog(
            self.parent,
            "Attach policy to group",
            "policy",
            "attach",
            PolicyLister,
            self.do_attach,
            caller=self,
        )

    def create_group(self, _):
        SingleNameDialog(
            self.parent,
            "Create new group",
            self.do_create,
            label="Name:",
            what="name",
            subject="group",
            default="",
            caller=self,
            accepted_inputs=Commons.alphanum_with_symbols("+=,.@_-"),
        )

    def do_create(self, name):
        Common.generic_api_call(
            "iam",
            "create_group",
            {"GroupName": name},
            "Create Group",
            "IAM",
            subcategory="Group",
            success_template="Creating group {GroupName}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def delete_group(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="group",
            resource_identifier=self.selection["name"],
            callback=self.do_delete,
            action_name="Delete",
        )

    def do_delete(self, **kwargs):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "delete_group",
            {"GroupName": self.selection["name"]},
            "Delete Group",
            "IAM",
            subcategory="Group",
            success_template="Deleting group {GroupName}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def remove_from_group(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="user",
            resource_identifier=self.user["name"],
            from_what="group",
            from_what_name=self.selection["name"],
            callback=self.do_remove,
            action_name="Remove",
        )

    def do_remove(self, **kwargs):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "remove_user_from_group",
            {"GroupName": self.selection["name"], "UserName": self.user["name"]},
            "Remove User From Group",
            "IAM",
            subcategory="Group",
            success_template="Removing user {UserName} from group {GroupName}",
            resource=self.user["name"],
        )
        self.refresh_data()

    def add_user_to_group(self, _):
        if self.selection is None:
            return
        SingleSelectorDialog(
            self.parent,
            "Add user to group",
            "user",
            "add",
            UserLister,
            self.do_add,
            caller=self,
        )

    def do_add(self, user):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "add_user_to_group",
            {"GroupName": self.selection["name"], "UserName": user},
            "Add User to Group",
            "IAM",
            subcategory="Group",
            success_template="Adding user {UserName} to group {GroupName}",
            resource=self.selection["name"],
        )

    def __init__(self, *args, user=None, **kwargs):
        self.resource_key = "iam"
        self.user = user
        self.list_method = "list_groups_for_user" if user is not None else "list_groups"
        self.item_path = ".Groups"
        self.list_kwargs = {"UserName": self.user["name"]} if user is not None else {}
        self.column_paths = {
            "name": ".GroupName",
            "id": ".GroupId",
            "path": ".Path",
        }
        self.imported_column_sizes = {
            "name": 25,
            "id": 15,
            "path": 40,
        }
        self.hidden_columns = {
            "arn": ".Arn",
        }
        self.describe_command = GroupDescriber.opener
        self.additional_commands = {
            "u": {
                "command": UserLister.opener,
                "selection_arg": "group",
                "tooltip": "List Group Users",
            },
            "m": {
                "command": PolicyLister.opener,
                "selection_arg": "group",
                "tooltip": "View Managed Policies",
            },
            "p": {
                "command": InlinePolicyLister.opener,
                "selection_arg": "data",
                "tooltip": "View Inline Policies",
                "kwargs": {
                    "datatype": "group",
                },
            },
        }

        self.imported_column_order = ["name", "id", "path"]
        self.sort_column = "name"
        super().__init__(*args, **kwargs)
        self.inliner = InlinePolicyAttacher(self, "group")
        self.add_hotkey("a", self.attach_policy, "Attach Policy")
        self.add_hotkey("c", self.create_group, "Create Group")
        self.add_hotkey("i", self.inliner.attach_inline_policy, "Put Inline Policy")
        if self.user is None:
            self.add_hotkey(ControlCodes.D, self.delete_group, "Delete Group")
            self.add_hotkey("j", self.add_user_to_group, "Add User")
        else:
            self.add_hotkey(ControlCodes.D, self.remove_from_group, "Remove from Group")


class GroupDescriber(Describer):
    prefix = "group_browser"
    title = "Group"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "iam"
        self.describe_method = "get_group"
        self.describe_kwarg_name = "GroupName"
        self.object_path = ".Group"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class PolicyLister(ResourceLister):
    prefix = "policy_list"
    title = "Policies"
    command_palette = ["policy", "policies"]

    def title_info(self):
        if self.user is not None:
            return "User: {0}".format(self.user["name"])
        elif self.group is not None:
            return "Group: {0}".format(self.group["name"])
        elif self.role is not None:
            return "Role: {0}".format(self.role["name"])
        else:
            return None

    def get_scope_tooltip(self):
        return "Change Scope ({0})".format(self.scope)

    def change_scope(self, *args):
        idx = self.scopes.index(self.scope)
        newidx = idx + 1
        if newidx >= len(self.scopes):
            newidx = 0
        self.scope = self.scopes[newidx]
        self.list_kwargs["Scope"] = self.scope
        Common.Session.set_message("Refreshing...", Common.color("message_info"))
        self.refresh_data()

    def create_policy(self, *args):
        creator = ListResourceDocumentCreator(
            "iam",
            "create_policy",
            None,
            as_json=["Tags"],
            initial_document={
                "PolicyName": "",
                "Path": "/",
                "PolicyDocument": {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Deny",
                            "Action": [""],
                            "Resource": "*",
                            "Condition": {},
                        }
                    ],
                },
                "Description": "",
                "Tags": [{"Key": "", "Value": ""}],
            },
        )
        creator.edit()
        self.refresh_data()

    def delete_policy(self, _):
        if "attached" not in self.selection or int(self.selection["attached"]) > 0:
            Common.Session.set_message(
                "Policy must be detached before deletion", Common.color("message_error")
            )
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="policy",
            resource_identifier=self.selection["name"],
            callback=self.do_delete,
            action_name="Delete",
        )

    def do_delete(self, **kwargs):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "delete_policy",
            {"PolicyArn": self.selection["arn"]},
            "Delete Policy",
            "IAM",
            subcategory="Policy",
            success_template="Deleting policy {PolicyArn}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def detach_policy(self, _):
        if self.selection is None:
            return
        if self.user is not None:
            dtype = "user"
            dname = self.user.name
        elif self.group is not None:
            dtype = "group"
            dname = self.group.name
        elif self.role is not None:
            dtype = "role"
            dname = self.role.name
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="policy",
            resource_identifier=self.selection["name"],
            from_what=dtype,
            from_what_name=dname,
            callback=self.do_detach,
            action_name="Detach",
        )

    def do_detach(self, **kwargs):
        if self.selection is None:
            return
        if self.user is not None:
            method = "detach_user_policy"
            arg_name = "UserName"
            arg_value = self.user.name
            resource = "user"
        elif self.group is not None:
            method = "detach_group_policy"
            arg_name = "GroupName"
            arg_value = self.group.name
            resource = "group"
        elif self.role is not None:
            method = "detach_role_policy"
            arg_name = "RoleName"
            arg_value = self.role.name
            resource = "role"
        else:
            return
        Common.generic_api_call(
            "iam",
            method,
            {"PolicyArn": self.selection["arn"], arg_name: arg_value},
            "Detach {0} Policy".format(resource.capitalize()),
            "IAM",
            subcategory="Policy",
            success_template="Detaching policy {PolicyArn} from {resource}",
            resource="{0} {1}".format(resource, arg_value),
        )
        self.refresh_data()

    def __init__(self, *args, user=None, group=None, role=None, **kwargs):
        isset = int(user is not None) + int(group is not None) + int(role is not None)
        if isset > 1:
            raise RuntimeError("Only one of user, group or role can be set.")
        self.user = user
        self.group = group
        self.role = role
        self.resource_key = "iam"
        self.scopes = ["All", "AWS", "Local"]
        self.scope = "All"
        if "selector_cb" in kwargs:
            self.primary_key = "arn"

        if user is not None:
            self.list_method = "list_attached_user_policies"
            self.list_kwargs = {"UserName": self.user["name"]}
            self.item_path = ".AttachedPolicies"
        elif group is not None:
            self.list_method = "list_attached_group_policies"
            self.list_kwargs = {"GroupName": self.group["name"]}
            self.item_path = ".AttachedPolicies"
        elif role is not None:
            self.list_method = "list_attached_role_policies"
            self.list_kwargs = {"RoleName": self.role["name"]}
            self.item_path = ".AttachedPolicies"
        else:
            self.list_method = "list_policies"

            self.list_kwargs = {"Scope": self.scope}
            self.item_path = ".Policies"

        self.column_paths = {
            "name": ".PolicyName",
            "arn": ".PolicyArn" if isset > 0 else ".Arn",
        }
        self.imported_column_sizes = {
            "name": 80,
            "arn": 80,
        }
        self.imported_column_order = ["name", "arn"]
        if isset == 0:
            self.column_paths["attached"] = ".AttachmentCount"
            self.imported_column_sizes["attached"] = 8
            self.imported_column_order.append("attached")

        self.describe_command = PolicyDescriber.opener

        self.sort_column = "name"
        super().__init__(*args, **kwargs)
        self.add_hotkey("c", self.create_policy, "Create new policy")
        if isset == 0:
            self.add_hotkey("s", self.change_scope, self.get_scope_tooltip)
            self.add_hotkey(ControlCodes.D, self.delete_policy, "Delete policy")
        else:
            self.add_hotkey(ControlCodes.D, self.detach_policy, "Detach policy")


class PolicyDescriber(Describer):
    prefix = "policy_browser"
    title = "Policy"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="arn", **kwargs
    ):
        self.resource_key = "iam"
        self.describe_method = "get_policy"
        self.describe_kwarg_name = "PolicyArn"
        self.object_path = ".Policy"
        self.additional_commands = {
            "d": {
                "command": PolicyVersionDescriber.opener,
                "data_arg": "policy",
                "tooltip": "View Policy Document",
                "kwargs": {"entry": None},
            }
        }
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class PolicyVersionDescriber(Describer):
    prefix = "policy_version_browser"
    title = "Policy Document"

    def title_info(self):
        return self.name

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        entry,
        *args,
        policy="",
        entry_key="arn",
        **kwargs
    ):
        self.resource_key = "iam"
        self.describe_method = "get_policy_version"
        self.object_path = ".PolicyVersion"
        self.arn = Common.Session.jq(".Arn").input(text=policy).first()
        self.name = Common.Session.jq(".PolicyName").input(text=policy).first()
        version = Common.Session.jq(".DefaultVersionId").input(text=policy).first()
        self.describe_kwargs = {"PolicyArn": self.arn, "VersionId": version}
        super().__init__(
            parent, alignment, dimensions, *args, entry=None, entry_key=None, **kwargs
        )


class InlinePolicyLister(ResourceLister):
    prefix = "inline_policy_list"
    title = "Inline Policies"

    def title_info(self):
        return "{0}: {1}".format(self.datatype.capitalize(), self.data["name"])

    def delete_policy(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="inline policy",
            resource_identifier=self.selection["name"],
            from_what=self.datatype,
            from_what_name=self.data["name"],
            callback=self.do_delete,
            action_name="Delete",
        )

    def do_delete(self, **kwargs):
        if self.selection is not None:
            try:
                self.delete(
                    PolicyName=self.selection["name"],
                    UserName=self.user["name"],
                )
                Common.Session.set_message(
                    "Deleting policy {0} from {2} {1}".format(
                        self.selection["name"], self.user["name"], self.datatype
                    ),
                    Common.color("message_success"),
                )
            except Exception as e:
                Common.Session.set_message(str(e), Common.color("message_error"))
        self.refresh_data()

    def edit_policy(self, _):
        if self.selection is None:
            return
        self.attacher.do_attach_inline(self.data["name"], selection_is_policy=True)

    def __init__(self, *args, datatype="", data=None, **kwargs):
        self.data = data
        self.datatype = datatype
        self.attacher = InlinePolicyAttacher(self, datatype)
        self.arg = "{0}Name".format(datatype.capitalize())
        self.resource_key = "iam"
        self.delete = getattr(
            Common.Session.service_provider("iam"), "delete_{0}_policy".format(datatype)
        )
        self.list_method = "list_{0}_policies".format(datatype)
        self.item_path = ".PolicyNames"
        self.list_kwargs = {self.arg: self.data["name"]}
        self.column_paths = {
            "name": ".",
        }
        self.imported_column_sizes = {
            "name": 25,
        }
        self.describe_command = InlinePolicyDescriber.opener
        self.describe_kwargs = {"datatype": datatype, "data": data}

        self.imported_column_order = ["name"]
        self.sort_column = "name"
        super().__init__(*args, **kwargs)
        self.add_hotkey(ControlCodes.D, self.delete_policy, "Delete Policy")
        self.add_hotkey("e", self.edit_policy, "Edit Policy")


class InlinePolicyDescriber(Describer):
    prefix = "user_inline_policy_browser"
    title = "Inline Policy"

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        entry,
        *args,
        entry_key="name",
        caller=None,
        datatype="",
        data=None,
        **kwargs
    ):
        self.resource_key = "iam"
        self.describe_method = "get_{0}_policy".format(datatype)
        self.arg = "{0}Name".format(datatype.capitalize())
        self.describe_kwarg_name = "PolicyName"
        self.describe_kwargs = {self.arg: data["name"]}
        self.object_path = "."
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            caller=caller,
            **kwargs
        )


class RoleLister(ResourceLister):
    prefix = "role_list"
    title = "Roles"
    command_palette = ["role", "roles"]

    def title_info(self):
        return (
            None
            if self.instanceprofile is None
            else "InstanceProfile: {0}".format(self.instanceprofile.name)
        )

    def create_role(self, _):
        creator = ListResourceDocumentCreator(
            "iam",
            "create_role",
            None,
            as_json=["Tags"],
            initial_document={
                "RoleName": "",
                "Path": "/",
                "PermissionsBoundary": "",
                "MaxSessionDuration": 3600,
                "AssumeRolePolicyDocument": {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Allow",
                            "Action": ["sts:AssumeRole"],
                            "Resource": "*",
                            "Condition": {},
                            "Principal": {},
                        }
                    ],
                },
                "Description": "",
                "Tags": [{"Key": "", "Value": ""}],
            },
        )
        creator.edit()
        self.refresh_data()

    def do_attach(self, policy_arn):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "attach_role_policy",
            {"PolicyArn": policy_arn, "RoleName": self.selection["name"]},
            "Attach Managed Role Policy",
            "IAM",
            subcategory="Role",
            success_template="Attaching policy {PolicyArn} to role {RoleName}",
            resource=self.selection["name"],
        )

    def attach_policy(self, _):
        if self.selection is None:
            return
        SingleSelectorDialog(
            self.parent,
            "Attach policy to role",
            "policy",
            "attach",
            PolicyLister,
            self.do_attach,
            caller=self,
        )

    def add_to_instance_profile(self, _):
        if self.selection is None:
            return
        SingleSelectorDialog(
            self.parent,
            "Add role to instance profile",
            "instance profile",
            "add",
            InstanceProfileLister,
            self.do_add,
            caller=self,
        )

    def do_add(self, ip):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "add_role_to_instance_profile",
            {"InstanceProfileName": ip, "RoleName": self.selection["name"]},
            "Add Role to Instance Profile",
            "IAM",
            subcategory="Role",
            success_template="Adding role {RoleName} to instance profile {InstanceProfileName}",
            resource=self.selection["name"],
        )

    def detach_from_instance_profile(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="role",
            resource_identifier=self.selection["name"],
            from_what="instance profile",
            from_what_name=self.instanceprofile["name"],
            callback=self.do_detach,
            action_name="Detach",
        )

    def do_detach(self, **kwargs):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "remove_role_from_instance_profile",
            {
                "InstanceProfileName": self.instanceprofile["name"],
                "RoleName": self.selection["name"],
            },
            "Remove Role from Instance Profile",
            "IAM",
            subcategory="Role",
            success_template="Removing role {RoleName} from instance profile {InstanceProfileName}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def delete_role(self, _):
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="role",
            resource_identifier=self.selection["name"],
            callback=self.do_delete,
            action_name="Delete",
        )

    def do_delete(self, **kwargs):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "delete_role",
            {"RoleName": self.selection["name"]},
            "Delete Role",
            "IAM",
            subcategory="Role",
            success_template="Deleting role {RoleName}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def __init__(self, *args, instanceprofile=None, **kwargs):
        self.instanceprofile = instanceprofile
        self.resource_key = "iam"
        self.list_method = (
            "list_roles" if self.instanceprofile is None else "get_instance_profile"
        )
        self.list_kwargs = (
            {}
            if self.instanceprofile is None
            else {"InstanceProfileName": self.instanceprofile["name"]}
        )
        self.item_path = (
            ".Roles" if self.instanceprofile is None else ".InstanceProfile.Roles"
        )
        self.column_paths = {
            "name": ".RoleName",
            "id": ".RoleId",
            "path": ".Path",
        }
        self.imported_column_sizes = {
            "name": 25,
            "id": 15,
            "path": 40,
        }
        self.hidden_columns = {
            "arn": ".Arn",
        }
        self.describe_command = RoleDescriber.opener
        self.additional_commands = {
            "m": {
                "command": PolicyLister.opener,
                "selection_arg": "role",
                "tooltip": "View Managed Policies",
            },
            "p": {
                "command": InlinePolicyLister.opener,
                "selection_arg": "data",
                "tooltip": "View Inline Policies",
                "kwargs": {
                    "datatype": "role",
                },
            },
            "n": {
                "command": InstanceProfileLister.opener,
                "selection_arg": "role",
                "tooltip": "View Instance Profiles",
            },
        }

        self.imported_column_order = ["name", "id", "path"]
        self.sort_column = "name"
        super().__init__(*args, **kwargs)
        self.inliner = InlinePolicyAttacher(self, "role")
        self.add_hotkey("a", self.attach_policy, "Attach Policy")
        self.add_hotkey("c", self.create_role, "Create Role")
        self.add_hotkey("i", self.inliner.attach_inline_policy, "Put Inline Policy")
        if self.instanceprofile is None:
            self.add_hotkey(
                "j", self.add_to_instance_profile, "Add to Instance Profile"
            )
            self.add_hotkey(
                ControlCodes.D,
                self.delete_role,
                "Delete Role",
            )
        else:
            self.add_hotkey(
                ControlCodes.D,
                self.detach_from_instance_profile,
                "Detach from Instance Profile",
            )


class RoleDescriber(Describer):
    prefix = "role_browser"
    title = "Role"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "iam"
        self.describe_method = "get_role"
        self.describe_kwarg_name = "RoleName"
        self.object_path = ".Role"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class InstanceProfileLister(ResourceLister):
    prefix = "instance_profile_list"
    title = "Instance Profiles"
    command_palette = ["instanceprofile", "instanceprofiles"]

    def title_info(self):
        return None if self.role is None else "Role: {0}".format(self.role["name"])

    def determine_role_count(self, instanceprofile):
        return str(len(instanceprofile["Roles"])) if "Roles" in instanceprofile else 0

    def add_role(self, _):
        if self.selection is None:
            return
        SingleSelectorDialog(
            self.parent,
            "Add role to instance profile",
            "role",
            "add",
            RoleLister,
            self.do_add,
            caller=self,
        )

    def do_add(self, role):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "add_role_to_instance_profile",
            {"InstanceProfileName": self.selection["name"], "RoleName": role},
            "Add Role to Instance Profile",
            "IAM",
            subcategory="Role",
            success_template="Adding role {RoleName} to instance profile {InstanceProfileName}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def remove_from_role(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="role",
            resource_identifier=self.role["name"],
            from_what="instance profile",
            from_what_name=self.selection["name"],
            callback=self.do_detach,
            action_name="Detach",
        )

    def do_detach(self, **kwargs):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "remove_role_from_instance_profile",
            {
                "RoleName": self.role["name"],
                "InstanceProfileName": self.selection["name"],
            },
            "Remove Role from Instance Profile",
            "IAM",
            subcategory="Instance Profile",
            success_template="Removing to role {RoleName} from instance profile {InstanceProfileName}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def delete_instance_profile(self, _):
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="instance profile",
            resource_identifier=self.selection["name"],
            callback=self.do_delete,
            action_name="Delete",
        )

    def do_delete(self, **kwargs):
        if self.selection is None:
            return
        Common.generic_api_call(
            "iam",
            "delete_instance_profile",
            {
                "InstanceProfileName": self.selection["name"],
            },
            "Delete Instance Profile",
            "IAM",
            subcategory="Instance Profile",
            success_template="Deleting instance profile {InstanceProfileName}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def __init__(self, *args, role=None, **kwargs):
        self.role = role
        self.resource_key = "iam"
        self.list_method = (
            "list_instance_profiles"
            if self.role is None
            else "list_instance_profiles_for_role"
        )
        self.list_kwargs = {} if self.role is None else {"RoleName": self.role["name"]}
        self.item_path = ".InstanceProfiles"
        self.column_paths = {
            "name": ".InstanceProfileName",
            "id": ".InstanceProfileId",
            "path": ".Path",
            "roles": self.determine_role_count,
        }
        self.imported_column_sizes = {
            "name": 25,
            "id": 15,
            "path": 40,
            "roles": 5,
        }
        self.hidden_columns = {
            "arn": ".Arn",
        }
        self.describe_command = InstanceProfileDescriber.opener
        self.open_command = "o"
        self.additional_commands = {
            # "i": {
            #    "tooltip": "View Instances",
            # },
            # "l": {
            #    "tooltip": "View Launch Configurations",
            # },
            "o": {
                "command": RoleLister.opener,
                "selection_arg": "instanceprofile",
                "tooltip": "View Roles",
            },
        }

        self.imported_column_order = ["name", "id", "path", "roles"]
        self.sort_column = "name"
        super().__init__(*args, **kwargs)
        if self.role is None:
            self.add_hotkey("j", self.add_role, "Attach Role")
            self.add_hotkey(
                ControlCodes.D, self.delete_instance_profile, "Delete Instance Profile"
            )
        else:
            self.add_hotkey(
                ControlCodes.D, self.remove_from_role, "Detach Instance Profile"
            )


class InstanceProfileDescriber(Describer):
    prefix = "instance_profile_browser"
    title = "Instance Profile"

    def populate_entry(self, *args, entry, entry_key, **kwargs):
        import sys

        print("{0} {1}".format(entry, entry_key), file=sys.stderr)
        super().populate_entry(*args, entry=entry, entry_key=entry_key, **kwargs)

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        from .arn import ARN

        self.resource_key = "iam"
        self.describe_method = "get_instance_profile"
        self.describe_kwarg_name = "InstanceProfileName"
        self.object_path = ".InstanceProfile"
        if entry_key == "id":  # id is an Arn from a single relation lister
            arn = ARN(entry["id"])
            entry_key = "name"
            entry["name"] = arn.resource_id_first
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



from .base_control import ResourceLister
from .common import Common
from .resource_ami import *
from .resource_asg import *
from .resource_beanstalk import *
from .resource_cfn import *
from .resource_cloudwatch import *
from .resource_dsg import *
from .resource_ebs import *
from .resource_ec2 import *
from .resource_ec2_class import *
from .resource_iam import *
from .resource_keypair import *
from .resource_lb import *
from .resource_lc import *
from .resource_listener import *
from .resource_r53 import *
from .resource_rds import *
from .resource_routing import *
from .resource_s3 import *
from .resource_sg import *
from .resource_sqs import *
from .resource_subnet import *
from .resource_tg import *
from .resource_vpc import *

Common.run_on_init(ResourceLister.register)



from .color import ColorBlackOnGold, ColorBlackOnOrange, ColorGold
from .common import Commons
from .control import Control


class DialogField:
    def __init__(self):
        self.highlightable = False
        self.centered = False
        Commons.UIInstance.dirty = True

    def input(self, inkey):
        return False

    def paint(self, x0, x1, y, selected=False):
        pass


class DialogFieldLabel(DialogField):
    def __init__(self, texts, default_color=ColorGold, centered=True):
        super().__init__()
        self.default_color = default_color
        if isinstance(texts, str):
            self.texts = [(texts, default_color)]
        else:
            self.texts = list(texts[:])
        self.centered = centered

    @property
    def text(self):
        return "".join([a[0] for a in self.texts])

    @text.setter
    def text(self, value):
        self.texts = [(value, self.default_color)]

    def add_text(self, text, color=None):
        if color is None:
            color = self.default_color
        self.texts.append((text, color))

    def paint(self, x0, x1, y, selected=False):
        x = x0
        if self.centered:
            textlen = 0
            for t in self.texts:
                textlen += len(t[0])
            w = x1 - x0 + 1
            x = int(w / 2) - int(textlen / 2) + x0 + 1
        for t in self.texts:
            Commons.UIInstance.print(t[0], xy=(x, y), color=t[1])
            x += len(t[0])


class DialogFieldButton(DialogField):
    def __init__(self, text, action, color=ColorGold, selected_color=ColorBlackOnGold):
        super().__init__()
        self.highlightable = True
        self.text = text
        self.color = color
        self.selected_color = selected_color
        self.centered = True
        self.action = action

    def input(self, inkey):
        if inkey.is_sequence and inkey.name == "KEY_ENTER":
            self.action()
            Commons.UIInstance.dirty = True
        return True

    def paint(self, x0, x1, y, selected=False):
        x = x0
        if self.centered:
            textlen = len(self.text) + 4
            w = x1 - x0 + 1
            x = int(w / 2) - int(textlen / 2) + x0
        Commons.UIInstance.print(
            "< {0} >".format(self.text),
            xy=(x, y),
            color=self.selected_color if selected else self.color,
        )


class DialogFieldCheckbox(DialogField):
    def __init__(
        self, label, checked=False, color=ColorGold, selected_color=ColorBlackOnGold
    ):
        super().__init__()
        self.highlightable = True
        self.checked = checked
        self.label = label
        self.color = color
        self.selected_color = selected_color
        self.char_unchecked = "☐"
        self.char_checked = "☑"

    def input(self, inkey):
        if inkey.is_sequence and inkey.name == "KEY_ENTER":
            self.checked = not self.checked
            Commons.UIInstance.dirty = True
        return True

    def paint(self, x0, x1, y, selected=False):
        x = x0
        text = "{0} {1}".format(
            self.char_checked if self.checked else self.char_unchecked, self.label
        )
        if self.centered:
            textlen = len(text)
            w = x1 - x0 + 1
            x = int(w / 2) - int(textlen / 2) + x0
        Commons.UIInstance.print(
            text, xy=(x, y), color=self.selected_color if selected else self.color
        )


class DialogFieldText(DialogField):
    def __init__(
        self,
        label,
        text="",
        color=ColorBlackOnOrange,
        selected_color=ColorBlackOnGold,
        label_color=ColorGold,
        label_min=0,
        password=False,
        accepted_inputs=None,
    ):
        super().__init__()
        self.highlightable = True
        self.left = 0
        self.text = text
        self.label = label
        self.color = color
        self.label_color = label_color
        self.label_min = label_min
        self.selected_color = selected_color
        self.centered = True
        self.password = password
        self.drawable = 0
        if accepted_inputs is None:
            self.accepted_inputs = Commons.TextfieldInputs
        else:
            self.accepted_inputs = accepted_inputs

    def input(self, inkey):
        if inkey.is_sequence:
            if inkey.name == "KEY_LEFT":
                if self.left > 0:
                    self.left -= 1
                Commons.UIInstance.dirty = True
                return True
            elif inkey.name == "KEY_RIGHT":
                self.left += 1
                Commons.UIInstance.dirty = True
                return True
            elif inkey.name == "KEY_HOME":
                self.left = 0
                Commons.UIInstance.dirty = True
                return True
            elif inkey.name == "KEY_END":
                self.left = len(self.text) - self.drawable
                if self.left < 0:
                    self.left = 0
                Commons.UIInstance.dirty = True
                return True
            elif inkey.name == "KEY_BACKSPACE" and len(self.text) > 0:
                self.text = self.text[:-1]
                Commons.UIInstance.dirty = True
                return True
            elif inkey.name == "KEY_DELETE":
                self.text = ""
                Commons.UIInstance.dirty = True
                return True
        elif inkey in self.accepted_inputs:
            self.text += inkey
            if len(self.text) > self.drawable:
                self.left = len(self.text) - self.drawable + 1
            Commons.UIInstance.dirty = True
            return True
        return False

    def paint(self, x0, x1, y, selected=False):
        x = x0
        Commons.UIInstance.print(self.label, xy=(x, y), color=self.label_color)
        x += max(len(self.label) + 1, self.label_min)
        space = x1 - x + 1
        self.drawable = space
        if self.left >= len(self.text):
            self.left = 0
        text = self.text[
            self.left : (
                self.left + space
                if len(self.text) > self.left + space
                else len(self.text)
            )
        ]
        if self.password:
            text = "*" * len(text)
        if len(text) < space:
            text += " " * (space - len(text))
        Commons.UIInstance.print(
            text, xy=(x, y), color=self.selected_color if selected else self.color
        )


class DialogControl(Control):
    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        *args,
        ok_action=None,
        cancel_action=None,
        background_color=None,
        confirm_text="Ok",
        cancel_text="Cancel",
        color=ColorGold,
        selected_color=ColorBlackOnGold,
        **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.highlighted = 0
        self.bookmark = 0
        self.background_color = background_color

        self.fields = []
        if confirm_text is not None:
            self.fields.append(
                DialogFieldButton(
                    confirm_text,
                    ok_action if ok_action is not None else self.nop,
                    color,
                    selected_color,
                )
            )
        if cancel_text is not None:
            self.fields.append(
                DialogFieldButton(
                    cancel_text,
                    cancel_action if cancel_action is not None else self.nop,
                    color,
                    selected_color,
                )
            )

    def nop(self):
        Commons.UIInstance.log("WARN: DialogControl NOP action executed.")

    def add_field(self, field):
        self.fields.insert(self.bookmark, field)
        self.bookmark += 1
        for i in range(len(self.fields)):
            elem = self.fields[i]
            if elem.highlightable:
                self.highlighted = i
                break

    def input(self, key):
        if key == "\t":
            orig = self.highlighted
            while True:
                self.highlighted += 1
                if self.highlighted >= len(self.fields):
                    self.highlighted -= len(self.fields)
                if self.fields[self.highlighted].highlightable:
                    break
                if self.highlighted == orig:
                    break
            Commons.UIInstance.dirty = True
            return True
        elif key.is_sequence and key.name == "KEY_BTAB":
            orig = self.highlighted
            while True:
                self.highlighted -= 1
                if self.highlighted < 0:
                    self.highlighted += len(self.fields)
                if self.fields[self.highlighted].highlightable:
                    break
                if self.highlighted == orig:
                    break
            Commons.UIInstance.dirty = True
            return True
        self.fields[self.highlighted].input(key)
        return True  # Modals should prevent input from being piped to others.

    def paint(self):
        c = self.corners()
        if self.background_color is not None:
            w = c[0][1] - c[0][0] + 1
            for row in range(c[1][0], c[1][1] + 1):
                Commons.UIInstance.print(
                    " " * w, xy=(c[0][0], row), color=self.background_color
                )
        super().paint()
        buttons = []
        indices = {}
        y = c[1][0] + (1 if self.border is None else 2)
        x0 = c[0][0] + (1 if self.border is None else 2)
        x1 = c[0][1] - (1 if self.border is None else 2)
        # Commons.UIInstance.log('Paint dialog: x0={0}, x1={1}'.format(x0, x1))
        idx = 0
        for item in self.fields:
            if isinstance(item, DialogFieldButton):
                buttons.append(item)
                indices[item] = idx
            else:
                item.paint(x0, x1, y, idx == self.highlighted)
                y += 1
            idx += 1
        y = c[1][1] - (1 if self.border is None else 2)
        if len(buttons) > 0:
            subdiv = float((x1 - x0 + 1) / len(buttons))
            w = int(subdiv)
            rem = subdiv - w
            remacc = 0
            for button in buttons:
                remacc += rem
                x1 = x0 + w
                if remacc > 0.5:
                    x1 += 1
                    remacc -= 1
                button.paint(x0, x1, y, indices[button] == self.highlighted)
                x0 = x1 + 1



class Palette:
    def __init__(self):
        pass

    def __call__(self, code, string, bold=False):
        return string


class Palette8Bit(Palette):
    def __call__(self, code, string, background=None, bold=False):
        out = ""
        if background is not None:
            out += "\033[48;5;" + str(background) + "m"
        out += "\033["
        if bold:
            out += "1;"
        out += "38;5;" + str(code) + "m" + string + "\033[0m"
        return out


class Color:
    def __init__(self, palette, code, background=None):
        self.palette = palette
        self.code = code
        self.background = background

    def __call__(self, string, bold=False):
        return self.palette(self.code, string, background=self.background, bold=bold)


ColorBlack = Color(Palette8Bit(), 0)
ColorWhite = Color(Palette8Bit(), 15)
ColorWhiteOnBlack = Color(Palette8Bit(), 15, background=0)
ColorBlackOnWhite = Color(Palette8Bit(), 0, background=15)
ColorGold = Color(Palette8Bit(), 220)
ColorGreen = Color(Palette8Bit(), 40)
ColorDarkGreen = Color(Palette8Bit(), 70)
ColorMagenta = Color(Palette8Bit(), 165)
ColorBlackOnGold = Color(Palette8Bit(), 0, background=220)
ColorBlackOnOrange = Color(Palette8Bit(), 0, background=208)
ColorBlackOnGray = Color(Palette8Bit(), 0, background=251)



import subprocess
from pathlib import Path

from .base_control import (
    DeleteResourceDialog,
    Describer,
    DialogFieldResourceListSelector,
    ListResourceDocumentCreator,
    ResourceLister,
    SingleNameDialog,
    SingleRelationLister,
)
from .common import Common, SessionAwareDialog
from .termui.alignment import CenterAnchor, Dimension
from .termui.control import Border
from .termui.dialog import DialogFieldCheckbox, DialogFieldLabel, DialogFieldText
from .termui.ui import ControlCodes


class EC2RelatedLister(SingleRelationLister):
    prefix = "ec2_related"
    title = "EC2 Instance related resources"

    def title_info(self):
        return self.instance_id

    def __init__(self, *args, ec2_entry=None, **kwargs):
        from .resource_ami import AMIDescriber
        from .resource_cfn import CFNDescriber
        from .resource_ebs import EBSDescriber
        from .resource_ec2_class import InstanceClassDescriber
        from .resource_iam import InstanceProfileDescriber
        from .resource_sg import SGDescriber
        from .resource_subnet import SubnetDescriber
        from .resource_vpc import VPCDescriber

        self.resource_key = "ec2"
        self.instance_id = ec2_entry["instance id"]
        self.describe_method = "describe_instances"
        self.describe_kwargs = {"InstanceIds": [self.instance_id]}
        self.object_path = ".Reservations[0].Instances[0]"
        self.resource_descriptors = [
            {
                "base_path": "[.SecurityGroups[].GroupId]",
                "type": "Security Group",
                "describer": SGDescriber.opener,
            },
            {
                "base_path": '[.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value]',
                "type": "Cloudformation Stack",
                "describer": CFNDescriber.opener,
            },
            {
                "base_path": "[.VpcId]",
                "type": "VPC",
                "describer": VPCDescriber.opener,
            },
            {
                "base_path": "[.NetworkInterfaces[].NetworkInterfaceId]",
                "type": "Network Interface",
            },
            {
                "base_path": "[.BlockDeviceMappings[].Ebs.VolumeId]",
                "type": "EBS Volume",
                "describer": EBSDescriber.opener,
            },
            {
                "base_path": "[.ImageId]",
                "type": "AMI",
                "describer": AMIDescriber.opener,
            },
            {
                "base_path": "[.KeyName]",
                "type": "Keypair",
            },
            {
                "base_path": "[.InstanceType]",
                "type": "Instance Type",
                "describer": InstanceClassDescriber.opener,
            },
            {
                "base_path": "[.SubnetId]",
                "type": "Subnet",
                "describer": SubnetDescriber.opener,
            },
            {
                "base_path": "[.IamInstanceProfile.Arn]",
                "type": "Instance Profile",
                "describer": InstanceProfileDescriber.opener,
            },
        ]
        super().__init__(*args, **kwargs)


class EC2ResourceLister(ResourceLister):
    prefix = "ec2_list"
    title = "EC2 Instances"
    command_palette = ["ec2", "instance"]

    def title_info(self):
        return self.title_info_data

    def __init__(self, *args, **kwargs):
        self.resource_key = "ec2"
        self.list_method = "describe_instances"
        self.title_info_data = None
        if "asg" in kwargs:
            self.list_kwargs = {
                "Filters": [
                    {
                        "Name": "tag:aws:autoscaling:groupName",
                        "Values": [kwargs["asg"]["name"]],
                    }
                ]
            }
            self.title_info_data = "ASG: {0}".format(kwargs["asg"]["name"])
        self.item_path = "[.Reservations[].Instances[]]"
        self.column_paths = {
            "instance id": ".InstanceId",
            "name": self.tag_finder_generator("Name"),
            "type": ".InstanceType",
            "vpc": ".VpcId",
            "public ip": ".PublicIpAddress",
            "key name": ".KeyName",
            "state": ".State.Name",
        }
        self.imported_column_sizes = {
            "instance id": 11,
            "name": 30,
            "type": 10,
            "vpc": 15,
            "public ip": 15,
            "key name": 30,
            "state": 10,
        }
        self.hidden_columns = {"image": ".ImageId"}
        self.describe_command = EC2Describer.opener
        self.imported_column_order = [
            "instance id",
            "name",
            "type",
            "vpc",
            "public ip",
            "key name",
            "state",
        ]
        self.sort_column = "instance id"
        self.primary_key = "instance id"
        super().__init__(*args, **kwargs)
        self.add_hotkey("s", self.ssh, "Open SSH")
        self.add_hotkey("l", self.new_instance, "Launch new instance")
        self.add_hotkey("m", self.metrics, "Metrics")
        self.add_hotkey("r", self.related, "View related resources")
        self.add_hotkey(ControlCodes.A, self.create_ami, "Create AMI from instance")
        self.add_hotkey(ControlCodes.S, self.stop_start_instance, "Stop/start instance")
        self.add_hotkey(ControlCodes.D, self.terminate_instance, "Terminate instance")

    def create_ami(self, _):
        if self.selection is None:
            return
        ListResourceDocumentCreator(
            "ec2",
            "create_image",
            None,
            initial_document={
                "BlockDeviceMappings": [
                    {
                        "DeviceName": "string",
                        "VirtualName": "string",
                        "Ebs": {
                            "DeleteOnTermination": False,
                            "Iops": 123,
                            "SnapshotId": "string",
                            "VolumeSize": 123,
                            "VolumeType": "gp2",
                            "KmsKeyId": "string",
                            "Throughput": 123,
                            "OutpostArn": "string",
                            "Encrypted": False,
                        },
                        "NoDevice": "string",
                    },
                ],
                "Description": "string",
                "Name": "string",
                "NoReboot": True,
                "TagSpecifications": [
                    {
                        "ResourceType": "capacity-reservation",
                        "Tags": [
                            {"Key": "string", "Value": "string"},
                        ],
                    },
                ],
            },
            as_json=["BlockDeviceMappings", "TagSpecifications"],
            static_fields={"InstanceId": self.selection["instance id"]},
        ).edit()

    def stop_start_instance(self, _):
        if self.selection is not None:
            if self.selection["state"] == "running":
                DeleteResourceDialog(
                    self.parent,
                    CenterAnchor(0, 0),
                    Dimension("80%|40", "10"),
                    caller=self,
                    resource_type="EC2 Instance",
                    resource_identifier=self.selection["instance id"],
                    callback=self.do_stop,
                    undoable=True,
                    action_name="Stop",
                )
            elif self.selection["state"] == "stopped":
                DeleteResourceDialog(
                    self.parent,
                    CenterAnchor(0, 0),
                    Dimension("80%|40", "10"),
                    caller=self,
                    resource_type="EC2 Instance",
                    resource_identifier=self.selection["instance id"],
                    callback=self.do_start,
                    undoable=True,
                    action_name="Start",
                )
            else:
                Common.Session.set_message(
                    "Instance is not in running or stopped state.",
                    Common.color("message_info"),
                )

    def terminate_instance(self, _):
        if self.selection is not None:
            DeleteResourceDialog(
                self.parent,
                CenterAnchor(0, 0),
                Dimension("80%|40", "10"),
                caller=self,
                resource_type="EC2 Instance",
                resource_identifier=self.selection["instance id"],
                callback=self.do_terminate,
                action_name="Terminate",
            )

    def metrics(self, _):
        from .resource_cloudwatch import MetricLister

        if self.selection is not None:
            Common.Session.push_frame(
                MetricLister.opener(
                    metric_namespace="AWS/EC2",
                    dimension=("InstanceId", self.selection["instance id"]),
                )
            )

    def related(self, _):
        if self.selection is not None:
            Common.Session.push_frame(EC2RelatedLister.opener(ec2_entry=self.selection))

    def do_start(self, **kwargs):
        if self.selection is None:
            return
        api_kwargs = {
            "InstanceIds": [self.selection["instance id"]],
        }
        Common.generic_api_call(
            "ec2",
            "start_instances",
            api_kwargs,
            "Start instance",
            "EC2",
            subcategory="Instance",
            success_template="Starting instance {0}",
            resource=self.selection["instance id"],
        )
        self.refresh_data()

    def do_stop(self, **kwargs):
        if self.selection is None:
            return
        api_kwargs = {
            "InstanceIds": [self.selection["instance id"]],
        }
        Common.generic_api_call(
            "ec2",
            "stop_instances",
            api_kwargs,
            "Stop instance",
            "EC2",
            subcategory="Instance",
            success_template="Stopping instance {0}",
            resource=self.selection["instance id"],
        )
        self.refresh_data()

    def do_terminate(self, **kwargs):
        if self.selection is None:
            return
        api_kwargs = {
            "InstanceIds": [self.selection["instance id"]],
        }
        Common.generic_api_call(
            "ec2",
            "terminate_instances",
            api_kwargs,
            "Terminate instance",
            "EC2",
            subcategory="Instance",
            success_template="Terminating instance {0}",
            resource=self.selection["instance id"],
        )
        self.refresh_data()

    def new_instance(self, _):
        EC2LaunchDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "20"),
            caller=self,
            weight=-500,
        )

    def ssh(self, _):
        if self.selection is None:
            return
        key = Common.Session.ssh_key
        kpr = Common.generic_api_call(
            "ec2",
            "describe_key_pairs",
            {"KeyNames": self.selection["key name"]},
            "Describe Key Pair",
            "EC2",
            subcategory="Key Pair",
            resource=self.selection["instance id"],
        )
        if kpr["Success"]:
            kp = kpr["Response"]
            if len(kp["KeyPairs"]) > 0:
                assoc = Common.Session.get_keypair_association(
                    kp["KeyPairs"][0]["KeyPairId"]
                )
                if assoc != "":
                    key = assoc
        p = Path.home() / ".ssh" / key
        if not p.exists():
            Common.Session.set_message(
                "Selected SSH key does not exist", Common.color("message_error")
            )
            return
        if self.selection is not None:
            if self.selection["public ip"] == "":
                Common.Session.set_message(
                    "No public IP associated with instance",
                    Common.color("message_error"),
                )
            else:
                EC2SSHDialog(
                    self.parent,
                    CenterAnchor(0, 0),
                    Dimension("80%|40", "10"),
                    instance_entry=self.selection,
                    caller=self,
                    weight=-500,
                )


class EC2Describer(Describer):
    prefix = "ec2_browser"
    title = "EC2 Instance"

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        entry,
        *args,
        entry_key="instance id",
        **kwargs
    ):
        self.resource_key = "ec2"
        self.describe_method = "describe_instances"
        self.describe_kwarg_name = "InstanceIds"
        self.describe_kwarg_is_list = True
        self.object_path = ".Reservations[0].Instances[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class EC2SSHDialog(SessionAwareDialog):
    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        instance_entry=None,
        caller=None,
        *args,
        **kwargs
    ):
        kwargs["border"] = Border(
            Common.border("ec2_ssh", "default"),
            Common.color("ec2_ssh_modal_dialog_border", "modal_dialog_border"),
            "SSH to instance",
            Common.color(
                "ec2_ssh_modal_dialog_border_title", "modal_dialog_border_title"
            ),
            instance_entry["instance id"],
            Common.color(
                "ec2_ssh_modal_dialog_border_title_info",
                "modal_dialog_border_title_info",
            ),
        )
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.instance_id = instance_entry["instance id"]
        self.ip = instance_entry["public ip"]
        def_text = (
            Common.Configuration["default_ssh_usernames"][Common.Session.ssh_key]
            if Common.Session.ssh_key in Common.Configuration["default_ssh_usernames"]
            else ""
        )
        self.use_instance_connect = DialogFieldCheckbox(
            "Use EC2 Instance Connect",
            color=Common.color(
                "ec2_ssh_modal_dialog_textfield_label", "modal_dialog_textfield_label"
            ),
            selected_color=Common.color(
                "ec2_ssh_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
        )
        self.username_textfield = DialogFieldText(
            "SSH username",
            text=def_text,
            color=Common.color(
                "ec2_ssh_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "ec2_ssh_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "ec2_ssh_modal_dialog_textfield_label", "modal_dialog_textfield_label"
            ),
            label_min=16,
        )
        self.add_field(self.use_instance_connect)
        self.add_field(self.username_textfield)
        self.highlighted = 0
        self.caller = caller

    def accept_and_close(self):
        ph = Path.home() / ".ssh" / Common.Session.ssh_key
        ph_pub = Path.home() / ".ssh" / "{0}.pub".format(Common.Session.ssh_key)
        if self.use_instance_connect:
            if not ph_pub.exists():
                Common.Session.set_message(
                    'Public key "{0}.pub" does not exist in ~/.ssh'.format(
                        Common.Session.ssh_key
                    ),
                    Common.color("message_error"),
                )
            with ph_pub.open("r") as f:
                pubkey = f.read()
            api_kwargs = {
                "InstanceId": self.instance_id,
                "InstanceOSUser": self.username_textfield.text,
                "SSHPublicKey": pubkey,
            }
            resp = Common.generic_api_call(
                "ec2-instance-connect",
                "send_ssh_public_key",
                api_kwargs,
                "Instance Connect",
                "EC2",
                subcategory="Instance Connect",
                resource=self.instance_id,
            )
            if not resp["Success"]:
                self.close()
                return
        ssh_cmd = "{0}@{1}".format(self.username_textfield.text, self.ip)
        Common.info(
            "SSH to instance: {0}".format(self.ssh_cmd),
            "SSH to instance",
            "EC2",
            resource=self.instance_id,
        )
        ex = Common.Session.ui.unraw(
            subprocess.run,
            [
                "bash",
                "-c",
                "ssh -o StrictHostKeyChecking=no -i {0} {1}".format(
                    str(ph.resolve()), ssh_cmd
                ),
            ],
        )
        Common.Session.set_message(
            "ssh exited with code {0}".format(ex.returncode),
            Common.color("message_info"),
        )
        self.close()

    def close(self):
        self.parent.remove_block(self)


class EC2LaunchDialog(SessionAwareDialog):
    def __init__(self, parent, alignment, dimensions, caller=None, *args, **kwargs):
        from .resource_ami import AMIResourceLister
        from .resource_ec2_class import InstanceClassResourceLister
        from .resource_keypair import KeyPairResourceLister
        from .resource_sg import SGResourceLister
        from .resource_subnet import SubnetResourceLister

        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "Launch EC2 instance",
            Common.color("modal_dialog_border_title"),
        )
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.add_field(DialogFieldLabel("Enter AWS instance details"))
        self.error_label = DialogFieldLabel(
            "", default_color=Common.color("modal_dialog_error")
        )
        self.add_field(self.error_label)
        self.add_field(DialogFieldLabel(""))
        self.name_field = DialogFieldText(
            "Name:",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.name_field)
        self.instance_type_field = DialogFieldResourceListSelector(
            InstanceClassResourceLister,
            "Instance type:",
            "t2.micro",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.instance_type_field)
        self.image_field = DialogFieldResourceListSelector(
            AMIResourceLister,
            "AMI:",
            "",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.image_field)
        self.keypair_field = DialogFieldResourceListSelector(
            KeyPairResourceLister,
            "Keypair:",
            "",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
            primary_key="name",
        )
        self.add_field(self.keypair_field)
        self.subnet_field = DialogFieldResourceListSelector(
            SubnetResourceLister,
            "Subnet:",
            "",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.subnet_field)
        self.secgroup_field = DialogFieldResourceListSelector(
            SGResourceLister,
            "Security group:",
            "",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.secgroup_field)
        self.caller = caller

    def accept_and_close(self):
        if self.name_field.text == "":
            self.error_label.text = "Name cannot be blank."
            return
        if self.instance_type_field.text == "":
            self.error_label.text = "Instance type cannot be blank."
            return
        if self.image_field.text == "":
            self.error_label.text = "AMI cannot be blank."
            return
        if self.keypair_field.text == "":
            self.error_label.text = "Key pair cannot be blank."
            return

        api_kwargs = {
            "ImageId": self.image_field.text,
            "InstanceType": self.instance_type_field.text,
            "KeyName": self.keypair_field.text,
            "MinCount": 1,
            "MaxCount": 1,
            "TagSpecifications": [
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": "Name", "Value": self.name_field.text}],
                },
            ],
        }
        if self.subnet_field.text != "":
            api_kwargs["SubnetId"] = self.subnet_field.text
            if self.secgroup_field.text != "":
                sgr = Common.generic_api_call(
                    "ec2",
                    "describe_security_groups",
                    {"GroupIds": self.secgroup_field.text},
                    "Describe Security Group",
                    "EC2",
                    subcategory="Security Group",
                    resource=self.secgroup_field.text,
                )
                if not sgr["Success"]:
                    return
                sg = sgr["Response"]

                snr = Common.generic_api_call(
                    "ec2",
                    "describe_subnets",
                    {"SubnetIds": self.subnet_field.text},
                    "Describe Subnet",
                    "EC2",
                    subcategory="Subnet",
                    resource=self.subnet_field.text,
                )
                if not snr["Success"]:
                    return
                sn = snr["Response"]

                if sg["VpcId"] == sn["VpcId"]:
                    api_kwargs["SecurityGroupIds"] = [self.secgroup_field.text]
                else:
                    Common.Session.set_message(
                        "Security group and subnet belong to different VPCs.",
                        Common.color("message_error"),
                    )
                    return
        Common.generic_api_call(
            "ec2",
            "run_instances",
            api_kwargs,
            "Launch Instance",
            "EC2",
            subcategory="Instance",
            success_template="Instance {0} is launching",
        )

        super().accept_and_close()

    def close(self):
        if self.caller is not None:
            self.caller.refresh_data()
        super().close()



from .termui.common import Commons
from .termui.control import Control


class Filterer(Control):
    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        session,
        color,
        symbol_color,
        autocomplete_color,
        inactive_color,
        *args,
        **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.session = session
        self.color = color
        self.symbol_color = symbol_color
        self.autocomplete_color = autocomplete_color
        self.inactive_color = inactive_color
        self.paused = False
        self.session.filterer = self
        self.accepted_input = Commons.TextfieldInputs
        self.text = ""

    def input(self, inkey):
        if self.paused and not (inkey.is_sequence and inkey.name == "KEY_ESCAPE"):
            return False
        if inkey.is_sequence:
            if inkey.name == "KEY_ENTER":
                self.pause()
            elif inkey.name == "KEY_ESCAPE":
                self.session.resource_main.filter = None
                self.close()
            elif inkey.name == "KEY_BACKSPACE":
                if len(self.text) > 0:
                    self.text = self.text[:-1]
                self.session.resource_main.filter = self.text
            elif inkey.name == "KEY_DELETE":
                self.text = ""
                self.session.resource_main.filter = self.text
            Commons.UIInstance.dirty = True
        elif str(inkey) in self.accepted_input:
            self.text += inkey
            self.session.resource_main.filter = self.text
            Commons.UIInstance.dirty = True
        return True

    def pause(self):
        self.paused = True

    def resume(self):
        self.paused = False

    def close(self):
        self.parent.remove_block(self)
        self.paused = True
        self.session.filterer = None

    def paint(self):
        super().paint()
        (x0, x1), (y0, y1) = self.inner
        Commons.UIInstance.print("/", xy=(x0 + 1, y0), color=self.symbol_color)
        space = (x1 - 1) - (x0 + 2) + 1
        t = self.text if len(self.text) <= space else self.text[:space]
        Commons.UIInstance.print(
            t,
            xy=(x0 + 2, y0),
            color=self.color if not self.paused else self.inactive_color,
        )


class Commander(Control):
    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        session,
        color,
        symbol_color,
        autocomplete_color,
        ok_color,
        error_color,
        *args,
        **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.session = session
        self.color = color
        self.symbol_color = symbol_color
        self.autocomplete_color = autocomplete_color
        self.ok_color = ok_color
        self.error_color = error_color
        self.session.commander = self
        self.accepted_input = Commons.TextfieldInputsAlphaNum
        self.text = ""
        self.options = session.commander_options

    def autocomplete(self):
        if len(self.text) == 0:
            return None
        opt_keys = self.options.keys()
        candidates = [cand for cand in opt_keys if cand.startswith(self.text.lower())]
        if len(candidates) > 0:
            candidates.sort(key=lambda x: len(x))
            return candidates[0]
        return None

    def accept_autocomplete(self):
        acp = self.autocomplete()
        if acp is not None:
            self.text = acp

    def input(self, inkey):
        if inkey.is_sequence:
            if inkey.name == "KEY_ENTER":
                self.accept_and_close()
            elif inkey.name == "KEY_ESCAPE":
                self.close()
            elif inkey.name == "KEY_BACKSPACE":
                if len(self.text) > 0:
                    self.text = self.text[:-1]
            elif inkey.name == "KEY_DELETE":
                self.text = ""
            elif inkey.name == "KEY_TAB":
                self.accept_autocomplete()
            Commons.UIInstance.dirty = True
        elif str(inkey) in self.accepted_input:
            self.text += inkey
            Commons.UIInstance.dirty = True
        return True

    def accept_and_close(self):
        l = self.text.lower()
        option = None
        if l in self.options:
            option = self.options[l]
        else:
            acp = self.autocomplete()
            if acp is not None:
                option = self.options[acp]
        if option is not None:
            self.session.replace_frame(option())
            self.close()

    def close(self):
        self.parent.remove_block(self)
        self.session.commander = None

    def paint(self):
        super().paint()
        (x0, x1), (y0, y1) = self.inner
        Commons.UIInstance.print(":", xy=(x0 + 1, y0), color=self.symbol_color)
        space = (x1 - 1) - (x0 + 2) + 1
        t = self.text if len(self.text) <= space else self.text[:space]
        acp = self.autocomplete()
        color = self.color
        if self.text.lower() in self.options:
            color = self.ok_color
        elif acp is None:
            color = self.error_color
        Commons.UIInstance.print(t, xy=(x0 + 2, y0), color=color)
        if acp is not None:
            acp = acp[len(self.text) :]
            if len(acp) > 0:
                Commons.UIInstance.print(
                    acp, xy=(x0 + 2 + len(self.text), y0), color=self.autocomplete_color
                )



from multiprocessing import Pipe, Process

import boto3
from botocore import config as botoconf
from botocore import exceptions

from .common import Common


class AWSSubprocessWrapper:
    def __init__(self, client):
        self.client = client

    def __getattr__(self, attr):
        if hasattr(self.client, attr):
            a = getattr(self.client, attr)
            if callable(a):
                return AWSSubprocessWrapper.SubprocessCallWrapper(a)
            return a
        raise AttributeError

    class SubprocessCallWrapper:
        def __init__(self, target):
            self.target = target

        def __call__(self, *args, **kwargs):
            own, remote = Pipe(False)
            p = Process(
                target=self.execute, args=[remote, *args], kwargs=kwargs, daemon=True
            )
            p.start()
            p.join()
            data = own.recv()
            if isinstance(data, Exception):
                raise data
            return data

        def execute(self, remote, *args, **kwargs):
            try:
                data = self.target(*args, **kwargs)
            except Exception as e:
                remote.send(e)
            remote.send(data)


class AWS:
    def __init__(self):
        Common.Session.context_update_hooks.append(self.idcaller)
        self.idcaller()

    def conf(self):
        return botoconf.Config(
            region_name=Common.Session.region,
            signature_version="v4",
        )

    def s3conf(self):
        return botoconf.Config(
            region_name=Common.Session.region,
            signature_version="s3v4",
        )

    def env_session(self):
        return boto3.Session()

    def __call__(self, service, keys=None):
        if service == "s3":
            config = self.s3conf()
        else:
            config = self.conf()
        client = boto3.client(
            service,
            aws_access_key_id=keys["access"]
            if keys is not None
            else Common.Configuration.keystore[Common.Session.context]["access"],
            aws_secret_access_key=keys["secret"]
            if keys is not None
            else Common.Configuration.keystore[Common.Session.context]["secret"],
            config=config,
        )
        # return AWSSubprocessWrapper(client)
        return client

    def whoami(self, keys=None):
        return self("sts", keys).get_caller_identity()

    def list_regions(self):
        return [
            region["RegionName"]
            for region in self("ec2").describe_regions(AllRegions=True)["Regions"]
        ]

    def idcaller(self):
        try:
            w = self.whoami()
            try:
                del Common.Session.info_display.special_colors["Account"]
                del Common.Session.info_display.special_colors["UserId"]
            except KeyError:
                pass
            Common.Session.info_display["UserId"] = w["UserId"]
            Common.Session.info_display["Account"] = w["Account"]
        except (exceptions.ClientError, KeyError) as e:
            Common.Session.info_display.special_colors["UserId"] = Common.color("error")
            Common.Session.info_display.special_colors["Account"] = Common.color(
                "error"
            )
            Common.Session.info_display["UserId"] = "ERROR"
            Common.Session.info_display["Account"] = "ERROR"
            Common.Session.ui.log("ERROR: From AWS API: {0}".format(e))



from .common import Commons


class Dimension:
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def __getitem__(self, key):
        if key not in (0, 1, "x", "y", "w", "h"):
            raise KeyError("Bad dimension key " + key)
        val = self()
        return val[0] if key in (0, "x", "w") else val[1]

    def _v(self, v, h):
        if isinstance(v, str):
            test = v.split("|")
            if len(test) > 1:
                return max([self._v(t, h) for t in test])
            test = v.split("-")
            if len(test) > 1:
                val = self._v(test[0], h)
                for i in range(1, len(test)):
                    val -= self._v(test[i], h)
                return val
            if v[-1] == "%":
                return int(h * float(v[:-1]) / 100) - 1
            else:
                return int(v)
        else:
            return int(v)

    def __call__(self):
        dim = Commons.UIInstance.dim
        w = self._v(self.width, dim[0])
        h = self._v(self.height, dim[1])
        return (w, h)


class TopLeftAnchor:
    def __init__(self, left, top):
        self.top = top
        self.left = left

    def anchor(self, parent):
        if parent is None:
            return (self.left, self.top)
        else:
            tlp = parent.topleft()
            return (tlp[0] + self.left, tlp[1] + self.top)

    def topleft(self, dim, parent):
        return self.anchor(parent)


class BottomLeftAnchor:
    def __init__(self, left, bottom):
        self.bottom = bottom
        self.left = left

    def anchor(self, parent):
        if parent is None:
            return (self.left, self.bottom)
        else:
            tlp = parent.topleft()
            brp = parent.bottomright()
            return (tlp[0] + self.left, brp[1] - self.bottom)

    def topleft(self, dim, parent):
        a = self.anchor(parent)
        return (a[0], a[1] - dim[1] + 1)


class TopRightAnchor:
    def __init__(self, right, top):
        self.top = top
        self.right = right

    def anchor(self, parent):
        if parent is None:
            return (self.right, self.top)
        else:
            tlp = parent.topleft()
            brp = parent.bottomright()
            return (brp[0] - self.right, tlp[1] + self.top)

    def topleft(self, dim, parent):
        a = self.anchor(parent)
        return (a[0] - dim[0] + 1, a[1])


class CenterAnchor:
    def __init__(self, xoffset, yoffset):
        self.xoffset = xoffset
        self.yoffset = yoffset

    def anchor(self, parent):
        tlp = (0, 0)
        dim = Commons.UIInstance.dim
        if parent is not None:
            tlp = parent.topleft()
            dim = parent.dimensions()
        return (
            int(dim[0] / 2) + self.xoffset + tlp[0],
            int(dim[1] / 2) + self.yoffset + tlp[1],
        )

    def topleft(self, dim, parent):
        anchor = self.anchor(parent)
        return (anchor[0] - int(dim[0] / 2), anchor[1] - int(dim[1] / 2))



from .base_control import Describer, MultiLister, ResourceLister
from .common import Common


class SGResourceLister(ResourceLister):
    prefix = "sg_list"
    title = "Security Groups"
    command_palette = ["sg", "securitygroup"]

    def __init__(self, *args, **kwargs):
        self.resource_key = "ec2"
        self.list_method = "describe_security_groups"
        self.item_path = ".SecurityGroups"
        self.column_paths = {
            "group id": ".GroupId",
            "name": ".GroupName",
            "vpc": ".VpcId",
            "ingress rules": self.determine_ingress_rules,
            "egress rules": self.determine_egress_rules,
        }
        self.imported_column_sizes = {
            "group id": 30,
            "name": 30,
            "vpc": 15,
            "ingress rules": 20,
            "egress rules": 20,
        }
        self.describe_command = SGDescriber.opener
        self.open_command = SGRelated.opener
        self.open_selection_arg = "compare_value"

        self.imported_column_order = [
            "group id",
            "name",
            "vpc",
            "ingress rules",
            "egress rules",
        ]
        self.sort_column = "name"
        self.primary_key = "group id"
        super().__init__(*args, **kwargs)
        self.add_hotkey("i", self.ingress, "View ingress rules")
        self.add_hotkey("e", self.egress, "View egress rules")

    def determine_ingress_rules(self, sg):
        return len(sg["IpPermissions"])

    def determine_egress_rules(self, sg):
        return len(sg["IpPermissionsEgress"])

    def ingress(self, *args):
        if self.selection is not None:
            Common.Session.push_frame(SGRuleLister.opener(sg_entry=self.selection))

    def egress(self, *args):
        if self.selection is not None:
            Common.Session.push_frame(
                SGRuleLister.opener(sg_entry=self.selection, egress=True)
            )


class SGRuleLister(ResourceLister):
    prefix = "sg_rule_list"
    title = "SG Rules"
    well_known = {
        20: "ftp",
        22: "ssh",
        23: "telnet",
        25: "smtp",
        80: "http",
        110: "pop3",
        220: "imap",
        443: "https",
        989: "ftps",
        1437: "dashboard-agent",
        1443: "mssql",
        3306: "mysql",
        3389: "rdp",
        5432: "pgsql",
        6379: "redis",
        8983: "solr",
        9200: "es",
    }

    def title_info(self):
        return "{0}: {1}".format(
            "Egress" if self.egress else "Ingress", self.sg_entry["group id"]
        )

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        sg_entry=None,
        egress=False,
        *args,
        **kwargs
    ):
        self.resource_key = "ec2"
        self.list_method = "describe_security_groups"
        if sg_entry is None:
            raise ValueError("sg_entry is required")
        self.sg_entry = sg_entry
        self.egress = egress
        self.list_kwargs = {"GroupIds": [self.sg_entry["group id"]]}
        self.item_path = ".SecurityGroups[0].IpPermissions{0}".format(
            "Egress" if egress else ""
        )
        self.column_paths = {
            "protocol": self.determine_protocol,
            "name": self.determine_name,
            "ip addresses": self.determine_ips,
            "security groups": self.determine_sgs,
            "port range": self.determine_port_range,
        }
        self.imported_column_sizes = {
            "name": 10,
            "protocol": 5,
            "ip addresses": 40,
            "security groups": 50,
            "port range": 20,
        }
        self.imported_column_order = [
            "name",
            "protocol",
            "ip addresses",
            "security groups",
            "port range",
        ]
        self.sort_column = "port range"
        super().__init__(parent, alignment, dimensions, *args, **kwargs)

    def determine_name(self, rule):
        if rule["IpProtocol"] == "-1":
            return "<all>"
        if rule["FromPort"] != rule["ToPort"]:
            return ""
        if rule["FromPort"] not in SGRuleLister.well_known:
            return ""
        return SGRuleLister.well_known[rule["FromPort"]]

    def determine_protocol(self, rule):
        if rule["IpProtocol"] == "-1" or rule["IpProtocol"] == -1:
            return "<all>"
        return rule["IpProtocol"]

    def determine_ips(self, rule):
        if "IpRanges" not in rule:
            return ""
        return ",".join(cidrs["CidrIp"] for cidrs in rule["IpRanges"])

    def determine_sgs(self, rule):
        if "UserIdGroupPairs" not in rule:
            return ""
        return ",".join(pair["GroupId"] for pair in rule["UserIdGroupPairs"])

    def determine_port_range(self, rule):
        if rule["IpProtocol"] == "-1":
            return "0-65535"
        if rule["FromPort"] != rule["ToPort"]:
            return "{0}-{1}".format(rule["FromPort"], rule["ToPort"])
        return str(rule["FromPort"])


class SGDescriber(Describer):
    prefix = "sg_browser"
    title = "Security Group"

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        entry,
        *args,
        entry_key="group id",
        **kwargs
    ):
        self.resource_key = "ec2"
        self.describe_method = "describe_security_groups"
        self.describe_kwarg_name = "GroupIds"
        self.describe_kwarg_is_list = True
        self.object_path = ".SecurityGroups[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class SGRelated(MultiLister):
    prefix = "sg_related"
    title = "Resources using Security Group"

    def title_info(self):
        return self.compare_value

    def __init__(self, *args, **kwargs):
        kwargs["compare_key"] = "group id"
        self.resource_descriptors = [
            {
                "resource_key": "ec2",
                "list_method": "describe_instances",
                "list_kwargs": {},
                "item_path": "[.Reservations[].Instances[]]",
                "column_paths": {
                    "type": lambda x: "EC2 Instance",
                    "id": ".InstanceId",
                    "name": self.tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": True,
                "compare_path": "[.SecurityGroups[].GroupId]",
            },
            {
                "resource_key": "rds",
                "list_method": "describe_db_instances",
                "list_kwargs": {},
                "item_path": ".DBInstances",
                "column_paths": {
                    "type": lambda x: "RDS Instance",
                    "id": ".DBInstanceIdentifier",
                    "name": ".Endpoint.Address",
                },
                "hidden_columns": {},
                "compare_as_list": True,
                "compare_path": "[.VpcSecurityGroups[].VpcSecurityGroupId]",
            },
        ]
        super().__init__(*args, **kwargs)



import shutil
import signal
import sys
import tempfile
import termios
import threading
import time
import tty
from pathlib import Path

from blessed import Terminal

from .alignment import Dimension, TopLeftAnchor
from .block import Block
from .common import Commons
from .control import Control
from .screen import Screen

FrameRate = 0.02


class ControlCodes:
    A = "\x01"
    B = "\x02"
    C = "\x03"
    D = "\x04"
    E = "\x05"
    F = "\x06"
    G = "\x07"
    K = "\x0b"
    L = "\x0c"
    M = "\x0d"
    N = "\x0e"
    O = "\x0f"
    P = "\x10"
    Q = "\x11"
    R = "\x12"
    S = "\x13"
    T = "\x14"
    U = "\x15"
    V = "\x16"
    W = "\x17"
    X = "\x18"
    Y = "\x19"
    Z = "\x1a"


class KeyDisplay(Control):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = "<no key>"

    def input(self, key):
        if key.is_sequence:
            self.text = "{0}".format((str(key), key.name, key.code))
        else:
            self.text = key

    def paint(self):
        super().paint()
        tl = self.topleft()
        Commons.UIInstance.print(self.text, xy=tl)


class UI:
    def __init__(self):
        if Commons.UIInstance is not None:
            raise RuntimeError("UI is a singleton.")
        Commons.UIInstance = self
        self.debug = 1
        self.top_block = Block(
            None, TopLeftAnchor(0, 0), Dimension("100%", "100%"), tag="top"
        )
        self.last_paint = time.time()
        self.last_size_refresh = time.time()
        self.term = Terminal()
        self._w = self.term.width
        self._h = self.term.height
        self.exit = False
        signal.signal(signal.SIGINT, self.set_exit)
        self.buf = Screen(self)
        self.tickers = []
        self.cache_dir = None
        self.mutex = threading.Lock()
        self.input_buffer = []
        self.dirty = False
        self.restore = None

    def filecache(self, obj, cb, *args, **kwargs):
        if self.cache_dir is None:
            self.cache_dir = tempfile.mkdtemp(prefix="awsc")
        objpath = Path(self.cache_dir) / obj
        if not objpath.exists():
            data = cb(str(objpath.resolve()), *args, **kwargs)
            return data
        try:
            with objpath.open("r") as f:
                return f.read()
        except UnicodeDecodeError:
            with objpath.open("rb") as f:
                return f.read()

    def print(self, out, xy=None, color=None, wrap=False, bounds=None, bold=False):
        self.dirty = True
        if bounds is None:
            bounds = ((0, self.w), (0, self.h))
        if xy is None:
            xy = (bounds[0][0], bounds[1][0])
        end = False
        if (xy[0] > bounds[0][1] and not wrap) or xy[1] > bounds[1][1]:
            return
        while not end:
            space = bounds[0][1] - xy[0]
            if len(out) > space:
                part = out[:space]
                out = out[space:]
            else:
                space = len(out)
                part = out
                out = ""
                end = True
            p = 0
            for i in range(xy[0], xy[0] + space):
                try:
                    char = self.buf[xy[1]][i]
                    char.value = part[p]
                    char.color = color
                    char.bold = bold
                    char.dirty = True
                except IndexError:
                    break
                p += 1
            xy = (bounds[0][0], xy[1] + 1)
            if not wrap or xy[1] >= bounds[1][1]:
                end = True

    def refresh_size(self):
        if self.last_paint - self.last_size_refresh > 1:
            self.last_size_refresh = self.last_paint
            self._w = self.term.width
            self._h = self.term.height

    @property
    def w(self):
        self.refresh_size()
        return self._w

    @property
    def h(self):
        self.refresh_size()
        return self._h

    @property
    def dim(self):
        self.refresh_size()
        return (self._w, self._h)

    def set_exit(self, *args, **kwargs):
        self.exit = True

    def before_paint(self):
        self.top_block.before_paint()

    def progress_bar_paint(self, perc):
        width = int(self.w * perc)
        with self.term.location(0, self.h - 1):
            perc_t = "{0:.2f}% ".format(perc * 100)
            print(
                "\033[38;5;220m" + perc_t + ((width - len(perc_t)) * "░") + "\033[0m",
                end="",
            )

    def check_one_key(self):
        return self.term.inkey(0.01, 0.01)

    def paint(self):
        self.buf.clear()
        self.top_block.paint()
        with self.term.location(0, 0):
            self.buf.output()
        self.last_paint = time.time()
        self.dirty = False

    def unraw(self, cmd, *args, **kwargs):
        # termios.tcsetattr(self.term._keyboard_fd, termios.TCSAFLUSH, self.restore)
        # pylint: disable=protected-access
        termios.tcsetattr(self.term._keyboard_fd, termios.TCSANOW, self.restore)
        # pylint: disable=protected-access
        self.term._line_buffered = True
        self.term.stream.write(self.term.normal_cursor)
        self.term.stream.flush()
        self.term.stream.write(self.term.exit_fullscreen)
        self.term.stream.flush()
        termios.tcflush(sys.stdin, termios.TCIOFLUSH)
        try:
            return cmd(*args, **kwargs)
        finally:
            self.term.stream.write(self.term.enter_fullscreen)
            self.term.stream.flush()
            self.term.stream.write(self.term.hide_cursor)
            self.term.stream.flush()
            # pylint: disable=protected-access
            tty.setraw(self.term._keyboard_fd, termios.TCSANOW)
            # pylint: disable=protected-access
            self.term._line_buffered = False

    def buffer_input(self):
        while True:
            key = self.term.inkey(None, 0.01)
            if key is not None and key != "":
                self.mutex.acquire()
                try:
                    self.input_buffer.append(key)
                finally:
                    self.mutex.release()

    def process_input_buffer(self, st):
        self.mutex.acquire()
        try:
            if len(self.input_buffer) > 0:
                key = self.input_buffer[0]
                self.input_buffer = self.input_buffer[1:]
                if key == "\x03":
                    raise KeyboardInterrupt
                self.top_block.input(key)
                ct = time.time()
                if ct - st > FrameRate - 0.02:
                    return
        finally:
            self.mutex.release()

    def main(self):
        # pylint: disable=protected-access
        self.restore = termios.tcgetattr(self.term._keyboard_fd)
        try:
            with self.term.fullscreen():
                with self.term.hidden_cursor():
                    with self.term.raw():
                        t = threading.Thread(target=self.buffer_input, daemon=True)
                        t.start()
                        while not self.exit:
                            # print(self.term.clear())
                            st = time.time()
                            self.process_input_buffer(st)
                            for func in self.tickers:
                                func()
                            self.before_paint()
                            if self.dirty or time.time() - self.last_paint > 3:
                                self.paint()
                            t2 = time.time()
                            d = FrameRate - (t2 - st)
                            if d > 0:
                                time.sleep(d)
        except KeyboardInterrupt:
            pass
        finally:
            if self.cache_dir is not None:
                shutil.rmtree(self.cache_dir)



from botocore import exceptions

from .common import Common, SessionAwareDialog
from .info import HotkeyDisplay
from .termui.alignment import CenterAnchor, Dimension, TopRightAnchor
from .termui.control import Border
from .termui.dialog import DialogFieldLabel, DialogFieldText
from .termui.list_control import ListControl, ListEntry


class DeleteContextDialog(SessionAwareDialog):
    def __init__(
        self, parent, alignment, dimensions, name="", caller=None, *args, **kwargs
    ):
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "Delete Context",
            Common.color("modal_dialog_border_title"),
        )
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.name = name
        self.add_field(
            DialogFieldLabel(
                [
                    ('Delete context "', Common.color("modal_dialog_label")),
                    (name, Common.color("modal_dialog_label_highlight")),
                    ('"?', Common.color("modal_dialog_label")),
                ]
            )
        )
        self.highlighted = 1
        self.caller = caller

    def input(self, inkey):
        if inkey.is_sequence and inkey.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(inkey)

    def accept_and_close(self):
        Common.Configuration.delete_context(self.name)
        self.close()

    def close(self):
        if self.caller is not None:
            self.caller.reload_contexts()
        super().close()


class ImportContextDialog(SessionAwareDialog):
    def __init__(self, parent, alignment, dimensions, caller=None, *args, **kwargs):
        self.accepts_inputs = True
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "New Context",
            Common.color("modal_dialog_border_title"),
        )
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.caller = caller
        self.add_field(DialogFieldLabel("Enter AWS context details"))
        self.error_label = DialogFieldLabel(
            "", default_color=Common.color("modal_dialog_error")
        )
        self.add_field(self.error_label)
        self.add_field(DialogFieldLabel(""))
        self.name_field = DialogFieldText(
            "Name:",
            label_min=12,
            color=Common.color(
                "context_add_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "context_add_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "context_add_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
        )
        self.add_field(self.name_field)
        creds = Common.Session.service_provider.env_session().get_credentials()
        if creds is None or not creds.access_key or not creds.secret_key:
            Common.Session.set_message(
                "No valid credentials in environment.", Common.color("message_error")
            )
            self.close()
            return
        self.access_key = creds.access_key
        self.secret_key = creds.secret_key
        self.access_key_field = DialogFieldLabel(
            [
                ("Access key: ", Common.color("generic")),
                (creds.access_key, Common.color("highlight")),
            ],
            centered=False,
        )
        self.add_field(self.access_key_field)
        self.secret_key_field = DialogFieldLabel(
            [
                ("Secret key: ", Common.color("generic")),
                ("*" * len(creds.secret_key), Common.color("highlight")),
            ],
            centered=False,
        )
        self.add_field(self.secret_key_field)

    def input(self, inkey):
        if not self.accepts_inputs:
            return True
        if inkey.is_sequence and inkey.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(inkey)

    def accept_and_close(self):
        if self.name_field.text == "":
            self.error_label.text = "Name cannot be blank."
            return
        try:
            sts = Common.Session.service_provider.whoami(
                keys={"access": self.access_key, "secret": self.secret_key}
            )
        except exceptions.ClientError as e:
            self.error_label.text = "Key verification failed"
            return

        self.accepts_inputs = False

        Common.Configuration.add_or_edit_context(
            self.name_field.text, sts["Account"], self.access_key, self.secret_key
        )
        self.close()

    def close(self):
        if self.caller is not None:
            self.caller.reload_contexts()
        super().close()


class AddContextDialog(SessionAwareDialog):
    def __init__(self, parent, alignment, dimensions, caller=None, *args, **kwargs):
        self.accepts_inputs = True
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "New Context",
            Common.color("modal_dialog_border_title"),
        )
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.add_field(DialogFieldLabel("Enter AWS context details"))
        self.error_label = DialogFieldLabel(
            "", default_color=Common.color("modal_dialog_error")
        )
        self.add_field(self.error_label)
        self.add_field(DialogFieldLabel(""))
        self.name_field = DialogFieldText(
            "Name:",
            label_min=16,
            color=Common.color(
                "context_add_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "context_add_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "context_add_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
        )
        self.add_field(self.name_field)
        self.access_key_field = DialogFieldText(
            "Access key:",
            label_min=16,
            color=Common.color(
                "context_add_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "context_add_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "context_add_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
        )
        self.add_field(self.access_key_field)
        self.secret_key_field = DialogFieldText(
            "Secret key:",
            label_min=16,
            color=Common.color(
                "context_add_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "context_add_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "context_add_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
            password=True,
        )
        self.add_field(self.secret_key_field)
        self.caller = caller

    def input(self, inkey):
        if not self.accepts_inputs:
            return True
        if inkey.is_sequence and inkey.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(inkey)

    def accept_and_close(self):
        if self.name_field.text == "":
            self.error_label.text = "Name cannot be blank."
            return
        if self.access_key_field.text == "":
            self.error_label.text = "Access key cannot be blank."
            return
        if self.secret_key_field.text == "":
            self.error_label.text = "Secret key cannot be blank."
            return
        try:
            sts = Common.Session.service_provider.whoami(
                keys={
                    "access": self.access_key_field.text,
                    "secret": self.secret_key_field.text,
                }
            )
        except exceptions.ClientError as e:
            self.error_label.text = "Key verification failed"
            return

        self.accepts_inputs = False

        Common.Configuration.add_or_edit_context(
            self.name_field.text,
            sts["Account"],
            self.access_key_field.text,
            self.secret_key_field.text,
        )
        self.close()

    def close(self):
        if self.caller is not None:
            self.caller.reload_contexts()
        super().close()


class ContextList(ListControl):
    def __init__(self, parent, alignment, dimensions, *args, **kwargs):
        super().__init__(
            parent,
            alignment,
            dimensions,
            color=Common.color("context_list_generic", "generic"),
            selection_color=Common.color("context_list_selection", "selection"),
            title_color=Common.color("context_list_heading", "column_title"),
            *args,
            **kwargs,
        )
        self.hotkey_display = HotkeyDisplay(
            self.parent,
            TopRightAnchor(1, 0),
            Dimension("33%|50", 8),
            self,
            session=Common.Session,
            highlight_color=Common.color("hotkey_display_title"),
            generic_color=Common.color("hotkey_display_value"),
            tag="context",
        )
        self.add_hotkey("a", self.add_new_context, "Add new context")
        self.add_hotkey("d", self.set_default_context, "Set as default")
        self.add_hotkey("i", self.import_context, "Import context")
        self.add_hotkey("KEY_ENTER", self.select_context, "Select context")
        self.add_hotkey("\x04", self.delete_selected_context, "Delete context")
        self.new_context_dialog = None
        self.add_column("account id", 12)
        self.add_column("default", 7)
        idx = 0

        self.reload_contexts()

    def reload_contexts(self):
        self.entries = []
        self.selected = 0
        idx = 0
        defa = 0
        for context, data in Common.Configuration["contexts"].items():
            if context == Common.Configuration["default_context"]:
                defa = idx
                d = "✓"
            else:
                d = " "
            self.add_entry(
                ListEntry(context, **{"account id": data["account_id"], "default": d})
            )

            idx += 1
        if self.selected >= len(self.entries) or self.selected < 0:
            self.selected = defa

    def add_new_context(self, _):
        AddContextDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "20"),
            caller=self,
            weight=-500,
        )

    def import_context(self, _):
        ImportContextDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "20"),
            caller=self,
            weight=-500,
        )

    def set_default_context(self, _):
        if self.selection is not None:
            Common.Configuration["default_context"] = self.selection.name
            Common.Configuration.write_config()
            self.reload_contexts()

    def select_context(self, _):
        if self.selection is not None:
            Common.Session.context = self.selection.name

    def delete_selected_context(self, _):
        if self.selection is not None:
            DeleteContextDialog(
                self.parent,
                CenterAnchor(0, 0),
                Dimension("80%|40", 10),
                name=self.selection.name,
                caller=self,
                weight=-500,
            )



from .common import Common, DefaultAnchor, DefaultBorder, DefaultDimension
from .info import HotkeyDisplay
from .termui.alignment import Dimension, TopRightAnchor
from .termui.list_control import ListControl, ListEntry


class CommanderOptionsLister(ListControl):
    title = "Commands"

    @classmethod
    def opener(cls, **kwargs):
        l = cls(
            Common.Session.ui.top_block,
            DefaultAnchor,
            DefaultDimension,
            weight=0,
            **kwargs,
        )
        l.border = DefaultBorder("help", "Commands", None)
        return [l, l.hotkey_display]

    def __init__(self, parent, alignment, dimensions, *args, **kwargs):
        super().__init__(
            parent,
            alignment,
            dimensions,
            color=Common.color("help_list_generic", "generic"),
            selection_color=Common.color("help_list_selection", "selection"),
            title_color=Common.color("help_list_heading", "column_title"),
            *args,
            **kwargs,
        )
        self.phony = {
            "ctx": "AWS Contexts",
            "context": "AWS Contexts",
            "region": "AWS Region",
            "ssh": "SSH Keys",
            "logs": "Logs",
        }
        self.hotkey_display = HotkeyDisplay(
            self.parent,
            TopRightAnchor(1, 0),
            Dimension("33%|50", 8),
            self,
            session=Common.Session,
            highlight_color=Common.color("hotkey_display_title"),
            generic_color=Common.color("hotkey_display_value"),
        )
        self.add_hotkey("KEY_ENTER", self.select_and_close, "Open")
        self.column_titles = {}
        self.column_order = []
        self.add_column("command", 12)
        self.add_column("resource", 48)
        self.reload()

    def reload(self, move=True):
        self.entries = []
        for cmd in Common.Session.commander_options:
            opt = Common.Session.commander_options[cmd]
            if hasattr(opt, "__self__"):
                le = ListEntry(cmd, command=cmd, resource=opt.__self__.title)
            elif cmd in self.phony:
                le = ListEntry(cmd, command=cmd, resource=self.phony[cmd])
            else:
                le = ListEntry(cmd, command=cmd, resource="")
            le.controller_data["fn"] = opt
            self.entries.append(le)
        self.entries.sort(key=lambda x: x["command"])

    def select_and_close(self, _):
        if self.selection is not None:
            Common.Session.replace_frame(self.selection.controller_data["fn"]())



import datetime

from .color import ColorBlackOnGold, ColorBlackOnOrange, ColorGold
from .common import Commons
from .control import Control


class ListControl(Control):
    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        *args,
        color=ColorGold,
        selection_color=ColorBlackOnGold,
        title_color=ColorBlackOnOrange,
        **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.entries = []
        self.hotkeys = {
            "KEY_DOWN": self.list_down,
            "KEY_UP": self.list_up,
            "KEY_PGUP": self.pageup,
            "KEY_PGDOWN": self.pagedown,
            "KEY_END": self.end,
            "KEY_HOME": self.home,
        }
        self.tooltips = {}
        self.color = color
        self.selection_color = selection_color
        self.title_color = title_color
        self.selected = 0
        self.column_titles = {"name": 30}
        self.column_order = ["name"]
        self.calculated = 0
        self._filter = None
        self.thread_share["clear"] = False
        self._cache = None
        self.thread_share["new_entries"] = []
        self.thread_share["finalize"] = False
        self.top = 0

    @property
    def filter(self):
        return self._filter

    @filter.setter
    def filter(self, value):
        Commons.UIInstance.dirty = True
        self._filter = value
        self.top = 0
        self.selected = 0
        self._cache = None

    def add_entry(self, entry):
        self.entries.append(entry)
        self.sort()
        self._cache = None
        Commons.UIInstance.dirty = True

    def add_hotkey(self, hotkey, action, tooltip=None):
        self.hotkeys[hotkey] = action
        if tooltip is not None:
            self.tooltips[hotkey] = tooltip
        Commons.UIInstance.dirty = True

    def input(self, key):
        inkey = str(key)
        if key.is_sequence:
            inkey = key.name
        else:
            inkey = inkey.lower()
        if inkey in self.hotkeys.keys():
            self.hotkeys[inkey](self)
            Commons.UIInstance.dirty = True
            return True
        return False

    @property
    def filtered(self):
        if self._cache is None:
            self._cache = [e for e in self.entries if e.matches_filter(self._filter)]
        return self._cache

    def list_down(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        if self.selected < len(self.filtered) - 1:
            self.selected += 1
        rows = self.rows
        if self.selected >= self.top + rows:
            self.top = max(0, self.selected - rows + 1)

    def list_up(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        if self.selected > 0:
            self.selected -= 1
        if self.selected < self.top:
            self.top = self.selected

    def pagedown(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        self.selected += self.rows
        self.selected = min(len(self.filtered) - 1, self.selected)
        rows = self.rows
        if self.selected >= self.top + rows:
            self.top = max(0, self.selected - rows + 1)

    def pageup(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        self.selected -= self.rows
        self.selected = max(0, self.selected)
        if self.selected < self.top:
            self.top = self.selected

    def home(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        self.selected = 0
        self.top = 0

    def end(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        self.selected = len(self.filtered) - 1
        self.top = max(0, self.selected - self.rows + 1)

    @property
    def selection(self):
        if len(self.filtered) > self.selected:
            return self.filtered[self.selected]
        return None

    @property
    def rows(self):
        c = self.corners()
        y = c[1][0] + (0 if self.border is None else 1)
        y1 = c[1][1] - (0 if self.border is None else 1)
        return y1 - y

    def add_column(self, column, min_size=0, index=None):
        if index is None:
            self.column_order.append(column)
        else:
            self.column_order.insert(index, column)
        self.column_titles[column] = min_size
        self.calculated = 0
        Commons.UIInstance.dirty = True

    def before_paint_critical(self):
        if self.thread_share["clear"]:
            Commons.UIInstance.dirty = True
            self.entries = []
            self.top = 0
            self.thread_share["clear"] = False
        if len(self.thread_share["new_entries"]) > 0:
            Commons.UIInstance.dirty = True
            self.handle_new_entries_critical(self.thread_share["new_entries"])
            self._cache = None
            self.thread_share["new_entries"] = []
        if self.thread_share["finalize"]:
            self.handle_finalization_critical()
            self.thread_share["finalize"] = False

    def handle_new_entries_critical(self, entries):
        self.entries.extend(entries)
        self.sort()

    def handle_finalization_critical(self):
        pass

    def before_paint(self):
        super().before_paint()
        self.mutex.acquire()
        try:
            self.before_paint_critical()
        finally:
            self.mutex.release()
        # if self.selected >= len(self.entries):
        #  self.selected = len(self.entries) - 1
        # if self.selected < 0:
        #  self.selected = 0

    def sort(self):
        pass

    def paint(self):
        win = self.w_in
        if win != self.calculated:
            vals = 0
            for k, v in self.column_titles.items():
                vals += v
            if vals > win:
                ratio = float(vals) / float(win)
                tot = 0
                for k in self.column_titles.keys():
                    self.column_titles[k] = int(float(self.column_titles[k]) / ratio)
                    tot += self.column_titles[k]
                rem = win - tot
                if "name" in self.column_titles:
                    self.column_titles["name"] += rem
                else:
                    self.column_titles[self.column_order[0]] += rem
            elif vals < win:
                diff = win - vals
                part = int(diff / len(self.column_titles))
                rem = diff - (part * len(self.column_titles))
                for k in self.column_titles.keys():
                    self.column_titles[k] += part
                if "name" in self.column_titles:
                    self.column_titles["name"] += rem
                else:
                    self.column_titles[self.column_order[0]] += rem
            self.calculated = win

        super().paint()
        c = self.corners()
        y = c[1][0] + (0 if self.border is None else 1)
        x = c[0][0] + (0 if self.border is None else 1)
        now = datetime.datetime.now()
        rows = self.rows
        for col in self.column_order:
            text = col
            m = self.column_titles[col]
            if len(col) > m:
                text = col[:m]
            elif len(col) < m:
                text = col + (int(m - len(col)) * " ")
            Commons.UIInstance.print(text.upper(), xy=(x, y), color=self.title_color)
            x += m
        y += 1
        for i in range(self.top, min(len(self.filtered), self.top + rows)):
            item = self.filtered[i]
            x = c[0][0] + (0 if self.border is None else 1)
            for col in self.column_order:
                text = item.columns[col]
                m = self.column_titles[col]
                if len(text) > m:
                    text = text[:m]
                elif len(text) < m:
                    text = text + ((m - len(text)) * " ")
                if i == self.selected:
                    if hasattr(
                        self, "update_selection_color"
                    ) and now - item.updated < datetime.timedelta(seconds=3):
                        color = self.update_selection_color
                    else:
                        color = self.selection_color
                else:
                    if hasattr(
                        self, "update_color"
                    ) and now - item.updated < datetime.timedelta(seconds=3):
                        color = self.update_color
                    else:
                        color = self.color
                Commons.UIInstance.print(text, xy=(x, y), color=color)
                x += m
            y += 1


class ListEntry:
    def __init__(self, name, controller_data=None, **kwargs):
        self.name = name
        self.columns = {"name": name}
        self.columns.update({k: str(v) for (k, v) in kwargs.items()})
        if controller_data is None:
            self.controller_data = {}
        else:
            self.controller_data = controller_data
        self.updated = datetime.datetime.now()

    def dict(self):
        return self.columns.copy()

    def __str__(self):
        return str(self.dict())

    def __getitem__(self, item):
        return self.columns[item]

    def __setitem__(self, item, value):
        self.columns[item] = value
        Commons.UIInstance.dirty = True

    def __contains__(self, item):
        return item in self.columns

    def matches_filter(self, filter):
        if filter is None:
            return True
        f = filter.lower()
        for k, column in self.columns.items():
            if f in column.lower():
                return True
        return False

    def mutate(self, other):
        up = False
        if self.name != other.name or len(self.columns) != len(other.columns):
            up = True
        else:
            for col in self.columns:
                if col not in other.columns or self.columns[col] != other.columns[col]:
                    up = True
                    break
        if not up:
            return
        self.updated = datetime.datetime.now()
        self.name = other.name
        self.columns = other.columns.copy()
        self.controller_data = other.controller_data



import datetime
import json

import yaml

from .base_control import GenericDescriber, datetime_hack
from .common import Common, DefaultAnchor, DefaultBorder, DefaultDimension
from .info import HotkeyDisplay
from .termui.alignment import Dimension, TopRightAnchor
from .termui.list_control import ListControl, ListEntry


class LogLister(ListControl):
    @classmethod
    def opener(cls, **kwargs):
        l = cls(
            Common.Session.ui.top_block,
            DefaultAnchor,
            DefaultDimension,
            weight=0,
            **kwargs,
        )
        l.border = DefaultBorder("log", "Logs", None)
        return [l, l.hotkey_display]

    @classmethod
    def selector(cls, cb, **kwargs):
        return cls.opener(**{"selector_cb": cb, **kwargs})

    def __init__(
        self, parent, alignment, dimensions, *args, selector_cb=None, **kwargs
    ):
        super().__init__(
            parent,
            alignment,
            dimensions,
            color=Common.color("ssh_key_list_generic", "generic"),
            selection_color=Common.color("ssh_key_list_selection", "selection"),
            title_color=Common.color("ssh_key_list_heading", "column_title"),
            *args,
            **kwargs,
        )
        self.selector_cb = selector_cb
        self.hotkey_display = HotkeyDisplay(
            self.parent,
            TopRightAnchor(1, 0),
            Dimension("33%|50", 8),
            self,
            session=Common.Session,
            highlight_color=Common.color("hotkey_display_title"),
            generic_color=Common.color("hotkey_display_value"),
        )
        self.add_hotkey("d", self.describe, "Show full entry")
        if selector_cb is not None:
            self.add_hotkey("KEY_ENTER", self.select_and_close, "Select and close")
        else:
            self.add_hotkey("KEY_ENTER", self.describe, "Show full entry")
        self.add_column("type", 12)
        self.add_column("category", 20)
        self.add_column("subcategory", 20)
        self.add_column("resource", 20)
        self.add_column("timestamp", 20)
        self.logholder = Common._logholder
        self.logholder.attach(self)

    def add_raw_entry(self, entry):
        self.add_entry(
            ListEntry(
                entry["summary"],
                **{
                    "category": entry["category"],
                    "subcategory": entry["subcategory"]
                    if "subcategory" in entry
                    and entry["subcategory"] is not None
                    and entry["subcategory"] != "null"
                    else "<n/a>",
                    "resource": entry["resource"]
                    if entry["resource"] is not None
                    else "",
                    "type": entry["type"],
                    "timestamp": datetime.datetime.utcfromtimestamp(
                        entry["timestamp"]
                    ).strftime("%Y-%m-%d %H:%M:%S"),
                    "raw_timestamp": entry["timestamp"],
                    "message": entry["message"],
                    "context": entry["context"] if "context" in entry else {},
                },
            )
        )

    def describe(self, _):
        if self.selection is not None:
            Common.Session.push_frame(LogViewer.opener(log_line=self.selection))

    def select_and_close(self, _):
        if self.selection is not None and self.selector_cb is not None:
            self.selector_cb(self.selection["name"])
            Common.Session.pop_frame()

    def sort(self):
        self.entries.sort(reverse=True, key=lambda x: x.columns["raw_timestamp"])
        self._cache = None

    def on_close(self):
        self.logholder.detach()


class LogViewer(GenericDescriber):
    def __init__(self, parent, alignment, dimensions, *args, log_line, **kwargs):
        columns = log_line.columns.copy()
        columns["context"] = json.loads(columns["context"])
        content = json.dumps(columns, default=datetime_hack, indent=2, sort_keys=True)
        super().__init__(
            parent,
            alignment,
            dimensions,
            describing="logs",
            content=content,
            **kwargs,
        )



import json
import os
import subprocess
import tempfile
import time
import urllib.request

import jq
from packaging import version

from .info import InfoDisplay, NeutralDialog
from .termui.alignment import BottomLeftAnchor, Dimension, TopLeftAnchor
from .termui.dialog import DialogFieldLabel
from .termui.ui import UI, ControlCodes
from .version import version as current_version


class Session:
    def jq(self, stmt):
        if stmt not in self.jqc:
            self.jqc[stmt] = jq.compile(stmt)
        return self.jqc[stmt]

    def __init__(self, config, highlight_color, generic_color, common):
        self.jqc = {}
        self.ui = UI()
        self.config = config
        self.context_update_hooks = []
        self.info_display = InfoDisplay(
            self.ui.top_block,
            TopLeftAnchor(1, 0),
            Dimension("50%", 8),
            highlight_color=highlight_color,
            generic_color=generic_color,
            info=[
                "AWSC Version",
                "Context",
                "Region",
                "UserId",
                "Account",
                "SSH Key",
                "Default SSH username",
            ],
            weight=-10,
        )
        self.global_hotkey_tooltips = {
            ":": "Command palette",
            "/": "Filter / search",
            ControlCodes.C: "Exit",
        }
        self.message_display = NeutralDialog(
            self.ui.top_block,
            BottomLeftAnchor(0, 0),
            Dimension("100%", 3),
            confirm_text=None,
            cancel_text=None,
            color=generic_color,
            weight=-10,
        )
        self._message_label = DialogFieldLabel("")
        self._message_label_l2 = DialogFieldLabel("")
        self.message_display.add_field(self._message_label)
        self.message_display.add_field(self._message_label_l2)
        self._context = None
        self.context = config["default_context"]
        self.region = config["default_region"]
        self.ssh_key = config["default_ssh_key"]
        self.stack = []
        self.stack_frame = []
        self.resource_main = None
        self.service_provider = None
        self.message_time = 0
        self.last_tick = time.time()
        self.ui.tickers.append(self.tick)

        self.filterer = None
        self.commander = None
        self.commander_options = {}
        self.common = common

        self.set_version_information()

    def set_version_information(self):
        update_character = "↑"
        version_str = current_version
        latest_tag = version.parse(current_version)
        latest_stable = latest_tag
        current_tag = latest_tag
        resp = urllib.request.urlopen("https://pypi.org/pypi/awsc/json")
        if resp.status == 200:
            package_info = resp.read()
            package_info_json = json.loads(package_info)
            for release in package_info_json["releases"].keys():
                release_tag = version.parse(release)
                if release_tag > latest_tag:
                    latest_tag = release_tag
                if not release_tag.is_prerelease and release_tag > latest_stable:
                    latest_stable = release_tag
            if latest_tag != current_tag:
                if latest_stable != current_tag:
                    version_str = "{current} ({update_character}{stable}, {update_character}?{development}".format(
                        current=str(current_tag),
                        stable=str(latest_stable),
                        development=str(latest_tag),
                        update_character=update_character,
                    )
                else:
                    version_str = "{current} ({update_character}?{development}".format(
                        current=str(current_tag),
                        development=str(latest_tag),
                        update_character=update_character,
                    )
            else:
                version_str = str(current_tag)

        self.info_display["AWSC Version"] = version_str

    def set_message(self, text, color):
        l1 = text
        l2 = ""
        if len(text) > self.ui.w:
            l1 = text[: self.ui.w]
            l2 = text[self.ui.w :]
            if len(l2) > self.ui.w:
                l2 = l2[: self.ui.w]
        self._message_label.texts = [(l1, color)]
        self._message_label_l2.texts = [(l2, color)]
        self.message_time = min(len(text) / 10.0, 5.0)
        self.last_tick = time.time()
        self.ui.dirty = True

    def replace_frame(self, new_frame, drop_stack=True):
        for elem in self.stack_frame:
            elem.parent.remove_block(elem)
        if drop_stack:
            self.stack = []
        self.resource_main = new_frame[0]
        self.stack_frame = new_frame[:]
        self.ui.dirty = True

    def push_frame(self, new_frame):
        self.stack.append(self.stack_frame[:])
        self.replace_frame(new_frame, drop_stack=False)
        if hasattr(new_frame[0], "add_hotkey"):
            if "KEY_ESCAPE" not in new_frame[0].tooltips:
                new_frame[0].add_hotkey("KEY_ESCAPE", self.pop_frame, "Back")
        self.ui.dirty = True

    def pop_frame(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        for elem in self.stack_frame:
            if hasattr(elem, "on_close"):
                elem.on_close()
        if len(self.stack) > 0:
            self.replace_frame(self.stack.pop(), drop_stack=False)
        for elem in self.stack_frame:
            elem.reparent()
        self.ui.dirty = True

    def extend_frame(self, control):
        self.stack_frame.append(control)
        self.ui.dirty = True

    def remove_from_frame(self, control):
        if hasattr(control, "on_close"):
            control.on_close()
        try:
            self.stack_frame.remove(control)
        except ValueError as e:
            for elem in reversed(self.stack):
                if control in elem:
                    elem.remove(control)
                    return
        self.ui.dirty = True

    def tick(self):
        delta = time.time() - self.last_tick
        self.last_tick += delta
        if self.message_time > 0:
            self.message_time -= delta
            if self.message_time <= 0:
                self._message_label.texts = []
                self._message_label_l2.texts = []
        if hasattr(self.resource_main, "auto_refresh"):
            self.resource_main.auto_refresh()

    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, value):
        if self._context == value:
            return
        self._context = value
        self.info_display["Context"] = value
        for elem in self.context_update_hooks:
            elem()

    @property
    def region(self):
        return self._region

    @region.setter
    def region(self, value):
        self._region = value
        self.info_display["Region"] = value

    @property
    def ssh_key(self):
        return self._ssh_key

    @ssh_key.setter
    def ssh_key(self, value):
        self._ssh_key = value
        self.info_display["SSH Key"] = value
        self.info_display["Default SSH username"] = (
            self.config["default_ssh_usernames"][value]
            if value in self.config["default_ssh_usernames"]
            else ""
        )

    def get_keypair_association(self, keypair_id):
        if keypair_id in self.config["keypair_associations"]:
            return self.config["keypair_associations"][keypair_id]
        return ""

    def set_keypair_association(self, keypair_id, key_name):
        self.config["keypair_associations"][keypair_id] = key_name
        self.config.write_config()

    def textedit(self, value):
        editor = self.config["editor_command"]
        temp = tempfile.NamedTemporaryFile("w", delete=False)
        tf = temp.name
        try:
            temp.write(value)
            temp.close()
            self.ui.unraw(subprocess.run, ["bash", "-c", editor.format(tf)])
            with open(tf, "r") as temp:
                return temp.read()
        finally:
            os.unlink(temp.name)



import pyperclip

# TODO: Figure out why this is a false positive.
# pylint: disable=no-name-in-module # I mean this is dangerous but I am getting a false positive here...
from pygments.lexers import JsonLexer
from pygments.token import Token

from .color import Color, ColorBlackOnGold, ColorGold, Palette8Bit
from .common import Commons
from .control import Control


class NullHighlighter:
    def __call__(self, browser, lines):
        return [[(line, browser.color)] for line in lines]


class JsonHighlighter:
    def _get_scheme_color(self, browser, color_name):
        try:
            c = browser.scheme["colors"][color_name]
        except KeyError:
            return browser.color
        return Color(Palette8Bit(), c["foreground"], background=c["background"])

    def __call__(self, browser, lines):
        import sys

        joined = "\n".join(lines)
        lexer = JsonLexer(tabsize=2, stripnl=False, stripall=False)
        tokens = lexer.get_tokens(joined)
        current_line = []
        result_lines = []
        for token in tokens:
            print("Token: {0}".format(token), file=sys.stderr)
            if token[0] in Token.Text:
                brk = token[1].split("\n")
                while len(brk) > 1:
                    current_line.append(
                        (
                            brk[0],
                            self._get_scheme_color(
                                browser,
                                "syntax_highlight_{0}".format(
                                    str(token[0]).lower().replace(".", "_")
                                ),
                            ),
                        )
                    )
                    result_lines.append(current_line)
                    current_line = []
                    brk = brk[1:]
                current_line.append(
                    (
                        brk[0],
                        self._get_scheme_color(
                            browser,
                            "syntax_highlight_{0}".format(
                                str(token[0]).lower().replace(".", "_")
                            ),
                        ),
                    )
                )
            else:
                current_line.append(
                    (
                        token[1],
                        self._get_scheme_color(
                            browser,
                            "syntax_highlight_{0}".format(
                                str(token[0]).lower().replace(".", "_")
                            ),
                        ),
                    )
                )
        result_lines.append(current_line)
        return result_lines


class TextBrowser(Control):
    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        *args,
        color=ColorGold,
        filtered_color=ColorBlackOnGold,
        scheme=None,
        syntax_highlighting=False,
        **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.entries = []
        self.hotkeys = {
            "KEY_DOWN": self.scroll_down,
            "KEY_UP": self.scroll_up,
            "KEY_LEFT": self.scroll_left,
            "KEY_RIGHT": self.scroll_right,
            "c": self.copy_contents,
            "w": self.toggle_wrap,
            "KEY_END": self.end,
            "KEY_HOME": self.home,
            "KEY_PGUP": self.pgup,
            "KEY_PGDOWN": self.pgdown,
        }
        self.tooltips = {"c": "Copy", "w": "Toggle Wrap"}
        self.scheme = scheme
        self.color = color
        self.filtered_color = filtered_color
        self.lines = []
        self.display_lines = []
        self._prefilter_lines = []
        self.top = 0
        self.left = 0
        self._filter_line = 0
        self.wrap = False
        self._filter = None
        self.filter_positions = {}
        self._current_filter_match = []
        self.syntax_highlighter = (
            JsonHighlighter() if syntax_highlighting else NullHighlighter()
        )

    def _raw(self, line_part):
        if isinstance(line_part, tuple):
            return line_part[0]
        return line_part

    def rawlines(self):
        return [self.raw(i) for i in range(len(self.lines))]

    def raw(self, line=None):
        if line is None:
            return "\n".join(self.rawlines())
        else:
            return "".join(
                [self._raw(self.lines[line][i]) for i in range(len(self.lines[line]))]
            )

    def add_text(self, text):
        add = text.split("\n")
        self.lines.extend(add)
        self._prefilter_lines.extend(add)
        self.display_lines = self.syntax_highlighter(self, self.lines)
        Commons.UIInstance.dirty = True

    def clear(self):
        Commons.UIInstance.dirty = True
        self.lines = []
        self.filter_positions = {}
        self._prefilter_lines = []
        self.filter = None
        self.top = 0
        self.left = 0

    @property
    def filter(self):
        return self._filter

    @filter.setter
    def filter(self, value):
        Commons.UIInstance.dirty = True
        refilter = False
        if self._filter != value:
            self._filter_line = -1
            self.filter_positions = {}
            self._current_filter_match = []
            refilter = True

        self._filter = value

        if value is None or value == "":
            return

        if refilter:
            for i in range(len(self.lines)):
                rawline = self.raw(i)
                ml = rawline.split(self._filter)
                pos = 0
                for j in range(1, len(ml)):
                    pos += len(ml[j - 1])
                    if i not in self.filter_positions:
                        self.filter_positions[i] = []
                    self.filter_positions[i].append(pos)
                if len(ml) > 1:
                    self._current_filter_match.append(i)

        for elem in self._current_filter_match:
            if elem > self._filter_line:
                self._filter_line = elem
                self.top = elem
                return
        if len(self._current_filter_match) > 0:
            self.top = self._current_filter_match[0]
            self._filter_line = self.top

    def add_hotkey(self, hotkey, action, tooltip=None):
        self.hotkeys[hotkey] = action
        if tooltip is not None:
            self.tooltips[hotkey] = tooltip
        Commons.UIInstance.dirty = True

    def input(self, key):
        inkey = str(key)
        if key.is_sequence:
            inkey = key.name
        else:
            inkey = inkey.lower()
        if inkey in self.hotkeys.keys():
            self.hotkeys[inkey](self)
            Commons.UIInstance.dirty = True
            return True
        return False

    def toggle_wrap(self, *args):
        self.wrap = not self.wrap
        if self.wrap:
            self.left = 0

    def scroll_down(self, *args):
        if self.top < len(self.lines) - 1:
            self.top += 1

    def scroll_up(self, *args):
        if self.top > 0:
            self.top -= 1

    def scroll_right(self, *args):
        if self.wrap:
            return
        self.left += 1

    def scroll_left(self, *args):
        if self.wrap:
            return
        if self.left > 0:
            self.left -= 1

    def end(self, *args):
        c = self.corners()
        y0 = c[1][0] + (0 if self.border is None else 1)
        y1 = c[1][1] - (0 if self.border is None else 1)
        h = y1 - y0 + 1
        limit = len(self.lines) - h if not self.wrap else len(self.lines) - 1
        self.top = max(limit, 0)

    def home(self, *args):
        self.top = 0

    def pgup(self, *args):
        c = self.corners()
        y0 = c[1][0] + (0 if self.border is None else 1)
        y1 = c[1][1] - (0 if self.border is None else 1)
        h = y1 - y0 + 1
        self.top = max(self.top - h, 0)

    def pgdown(self, *args):
        c = self.corners()
        y0 = c[1][0] + (0 if self.border is None else 1)
        y1 = c[1][1] - (0 if self.border is None else 1)
        h = y1 - y0 + 1
        limit = len(self.lines) - h if not self.wrap else len(self.lines) - 1
        self.top = max(min(limit, self.top + h), 0)

    def copy_contents(self, *args):
        pyperclip.copy(self.raw())

    def paint(self):
        import sys

        super().paint()
        print(
            "Lines ({0}): {1}".format(len(self.display_lines), self.display_lines),
            file=sys.stderr,
        )
        c = self.corners()
        y = c[1][0] + (0 if self.border is None else 1)
        y1 = c[1][1] - (0 if self.border is None else 1)
        x0 = c[0][0] + (0 if self.border is None else 1)
        x1 = c[0][1] - (0 if self.border is None else 1)
        for i in range(self.top, len(self.display_lines)):
            line = self.display_lines[i]
            skip_chars = self.left
            x = x0
            end = False
            pos = skip_chars
            next_filter_pos = -1
            next_filter_last_pos = -1
            nfp_idx = 0
            if i in self.filter_positions:
                for p in self.filter_positions[i]:
                    if p >= pos:
                        next_filter_pos = p
                        next_filter_last_pos = p + len(self._filter) - 1
                        break
                    elif p + len(self._filter) - 1 >= pos:
                        next_filter_pos = p
                        next_filter_last_pos = p + len(self._filter) - 1
                    else:
                        nfp_idx += 1
            for elem in line:
                buf = self._raw(elem)
                if skip_chars > len(buf):
                    skip_chars -= len(buf)
                    continue
                elif skip_chars > 0:
                    buf = buf[skip_chars:]
                    skip_chars = 0
                color = self.color if not isinstance(elem, tuple) else elem[1]

                while len(buf) > 0:
                    space = x1 - x + 1
                    text = buf[:space]
                    buf = buf[space:]
                    while len(text) > 0:
                        if (
                            next_filter_last_pos >= 0
                            and pos >= next_filter_pos
                            and pos <= next_filter_last_pos
                        ):
                            filt = next_filter_last_pos - pos + 1
                            if filt >= len(text):
                                Commons.UIInstance.print(
                                    text, xy=(x, y), color=self.filtered_color
                                )
                                x += len(text)
                                pos += len(text)
                                text = ""
                            else:
                                Commons.UIInstance.print(
                                    text[:filt], xy=(x, y), color=self.filtered_color
                                )
                                text = text[filt:]
                                x += filt
                                pos += filt
                                nfp_idx += 1
                                if len(self.filter_positions[i]) > nfp_idx:
                                    next_filter_pos = self.filter_positions[i][nfp_idx]
                                    next_filter_last_pos = (
                                        next_filter_pos + len(self._filter) - 1
                                    )
                                else:
                                    next_filter_pos = -1
                                    next_filter_last_pos = -1
                        elif next_filter_pos >= 0 and pos + len(text) > next_filter_pos:
                            unfilt = next_filter_pos - pos
                            if unfilt > 0:
                                Commons.UIInstance.print(
                                    text[:unfilt], xy=(x, y), color=color
                                )
                                text = text[unfilt:]
                                x += unfilt
                                pos += unfilt
                            filt = next_filter_last_pos - next_filter_pos + 1
                            if filt >= len(text):
                                Commons.UIInstance.print(
                                    text, xy=(x, y), color=self.filtered_color
                                )
                                x += len(text)
                                pos += len(text)
                                text = ""
                            else:
                                Commons.UIInstance.print(
                                    text[:filt], xy=(x, y), color=self.filtered_color
                                )
                                text = text[filt:]
                                x += filt
                                pos += filt
                                nfp_idx += 1
                                if len(self.filter_positions[i]) > nfp_idx:
                                    next_filter_pos = self.filter_positions[i][nfp_idx]
                                    next_filter_last_pos = (
                                        next_filter_pos + len(self._filter) - 1
                                    )
                                else:
                                    next_filter_pos = -1
                                    next_filter_last_pos = -1
                        else:
                            Commons.UIInstance.print(text, xy=(x, y), color=color)
                            pos += len(text)
                            x += len(text)
                            text = ""
                    if x > x1:
                        y += 1
                        x = x0
                        if not self.wrap:
                            end = True

                    if end or y > y1:
                        break
            if not end:
                y += 1
                x = x0
            if y > y1:
                break



import threading

from .block import Block
from .common import Commons


class BorderStyle:
    def __init__(self, chars=None):
        if chars is None:
            chars = ["-", "|", "-", "-", "-", "-"]
        self.chars = chars

    @property
    def horizontal(self):
        return self.chars[0]

    @property
    def vertical(self):
        return self.chars[1]

    @property
    def TL(self):
        return self.chars[2]

    @property
    def TR(self):
        return self.chars[3]

    @property
    def BL(self):
        return self.chars[4]

    @property
    def BR(self):
        return self.chars[5]


BorderStyleContinuous = BorderStyle(["─", "│", "┌", "┐", "└", "┘"])


class Border:
    def __init__(
        self,
        style=None,
        color=None,
        title=None,
        title_color=None,
        title_info=None,
        title_info_color=None,
    ):
        self.style = style
        self.color = color
        self.title = title
        self.title_color = title_color
        self.title_info = title_info
        self.title_info_color = title_info_color

    def paint(self, block):
        if self.style is None or self.color is None:
            return
        corners = block.corners()
        dimensions = block.dimensions()
        for i in range(corners[1][0], corners[1][1] + 1):
            if i == corners[1][0]:
                if self.title is None or self.title_color is None:
                    outer = (
                        self.style.TL
                        + self.style.horizontal * (dimensions[0] - 1)
                        + self.style.TR
                    )
                    Commons.UIInstance.print(
                        outer, xy=(corners[0][0], i), color=self.color
                    )
                else:
                    total_title = self.title
                    if self.title_info is not None:
                        total_title = "{0} ({1})".format(self.title, self.title_info)
                    cpos = int((dimensions[0] + 1) / 2)
                    spos = cpos - int(len(total_title) / 2) - 1
                    slen = len(total_title) + 2
                    outer = self.style.TL + self.style.horizontal * spos
                    l = len(outer)
                    Commons.UIInstance.print(
                        outer, xy=(corners[0][0], i), color=self.color
                    )
                    outer = " " + self.title + " "
                    Commons.UIInstance.print(
                        outer,
                        xy=(corners[0][0] + l, i),
                        color=self.title_color,
                        bold=True,
                    )
                    l += len(outer)
                    if self.title_info is not None:
                        Commons.UIInstance.print(
                            "(",
                            xy=(corners[0][0] + l, i),
                            color=self.title_color,
                            bold=True,
                        )
                        l += 1
                        Commons.UIInstance.print(
                            self.title_info,
                            xy=(corners[0][0] + l, i),
                            color=self.title_color
                            if self.title_info_color is None
                            else self.title_info_color,
                            bold=True,
                        )
                        l += len(self.title_info)
                        Commons.UIInstance.print(
                            ") ",
                            xy=(corners[0][0] + l, i),
                            color=self.title_color,
                            bold=True,
                        )
                        l += 2
                    outer = (
                        self.style.horizontal * (dimensions[0] - 1 - spos - slen)
                        + self.style.TR
                    )
                    Commons.UIInstance.print(
                        outer, xy=(corners[0][0] + l, i), color=self.color
                    )
            elif i == corners[1][1]:
                outer = (
                    self.style.BL
                    + self.style.horizontal * (dimensions[0] - 1)
                    + self.style.BR
                )
                Commons.UIInstance.print(outer, xy=(corners[0][0], i), color=self.color)
            else:
                outer = (
                    self.style.vertical
                    + " " * (dimensions[0] - 1)
                    + self.style.vertical
                )
                Commons.UIInstance.print(outer, xy=(corners[0][0], i), color=self.color)


BorderNone = Border()


class Control(Block):
    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        *args,
        weight=0,
        tag="default",
        border=None,
        **kwargs
    ):
        super().__init__(
            parent, alignment, dimensions, *args, weight=weight, tag=tag, **kwargs
        )
        self.border = border
        self.thread_share = {}
        self.mutex = threading.Lock()

    @property
    def border(self):
        return self._border

    @border.setter
    def border(self, value):
        self._border = value
        Commons.UIInstance.dirty = True

    @property
    def inner(self):
        c = self.corners()
        x0 = c[0][0] + (0 if self.border is None else 1)
        x1 = c[0][1] - (0 if self.border is None else 1)
        y0 = c[1][0] + (0 if self.border is None else 1)
        y1 = c[1][1] - (0 if self.border is None else 1)
        return ((x0, x1), (y0, y1))

    @property
    def w_in(self):
        return self.w if self.border is None else self.w - 2

    def paint(self):
        if self.border is not None:
            self.border.paint(self)
        super().paint()



from pathlib import Path

from .base_control import Describer, ResourceLister
from .common import Common, SessionAwareDialog
from .ssh import SSHList
from .termui.control import Border
from .termui.dialog import DialogFieldLabel, DialogFieldText


class KeyPairResourceLister(ResourceLister):
    prefix = "keypair_list"
    title = "EC2 Keypairs"
    command_palette = ["key", "keypair"]

    def __init__(self, *args, **kwargs):
        self.resource_key = "ec2"
        self.list_method = "describe_key_pairs"
        self.item_path = ".KeyPairs"
        self.column_paths = {
            "id": ".KeyPairId",
            "name": ".KeyName",
            "fingerprint": ".KeyFingerprint",
            "association": self.determine_keypair_association,
        }
        self.imported_column_sizes = {
            "id": 16,
            "name": 32,
            "fingerprint": 48,
            "association": 16,
        }
        self.describe_command = KeyPairDescriber.opener
        self.imported_column_order = ["id", "name", "fingerprint", "association"]
        self.primary_key = "id"
        self.sort_column = "id"
        super().__init__(*args, **kwargs)
        self.add_hotkey("a", self.set_keypair_association, "Associate with SSH key")
        self.add_hotkey("c", self.create_keypair, "Create new keypair")
        self.add_hotkey("x", self.remove_keypair_association, "Remove association")

    def determine_keypair_association(self, entry):
        return Common.Session.get_keypair_association(entry["KeyPairId"])

    def set_keypair_association(self, _):
        if self.selection is not None:
            Common.Session.push_frame(SSHList.selector(self.ssh_selector_callback))
            Common.Session.ui.dirty = True

    def ssh_selector_callback(self, val):
        if self.selection is not None:
            Common.Session.set_keypair_association(self.selection["id"], val)
            self.refresh_data()

    def create_keypair(self, _):
        KeyPairCreateDialog.opener(self)

    def remove_keypair_association(self, _):
        if self.selection is not None:
            Common.Session.set_keypair_association(self.selection["id"], "")
            self.refresh_data()


class KeyPairCreateDialog(SessionAwareDialog):
    def __init__(self, *args, caller=None, **kwargs):
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "Create new keypair",
            Common.color("modal_dialog_border_title"),
        )
        super().__init__(*args, caller=caller, **kwargs)
        self.add_field(DialogFieldLabel("Enter keypair details"))
        self.error_label = DialogFieldLabel(
            "", default_color=Common.color("modal_dialog_error")
        )
        self.add_field(self.error_label)
        self.name_field = DialogFieldText(
            "Name:",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.name_field)
        self.dotssh = Path.home() / ".ssh"
        self.add_field(
            DialogFieldLabel(
                "Key file will be saved in {0}".format(str(self.dotssh.resolve()))
            )
        )
        self.save_as_field = DialogFieldText(
            "Filename:",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.save_as_field)
        self.caller = caller

    def accept_and_close(self):
        if self.name_field.text == "":
            self.error_label.text = "Name cannot be blank."
            return
        if self.save_as_field.text == "":
            self.error_label.text = "Filename cannot be blank."
            return
        if "/" in self.save_as_field.text or "\\" in self.save_as_field.text:
            self.error_label.text = "Filename cannot contain path separators."
            return

        resps = Common.generic_api_call(
            "ec2",
            "create_key_pair",
            {"KeyName": self.name_field.text},
            "Create Keypair",
            "EC2",
            subcategory="Key Pair",
            success_template="Creating keypair {0}",
            resource=self.name_field.text,
        )
        if resps["Success"]:
            resp = resps["Response"]
            data = resp["KeyMaterial"]
            with (self.dotssh / self.save_as_field.text).open("w") as f:
                f.write(data)
            Common.Session.set_keypair_association(
                resp["KeyPairId"], self.save_as_field.text
            )

        super().accept_and_close()

    def close(self):
        if self.caller is not None:
            self.caller.refresh_data()
        super().close()


class KeyPairDescriber(Describer):
    prefix = "keypair_browser"
    title = "EC2 Keypair"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="id", **kwargs
    ):
        self.resource_key = "ec2"
        self.describe_method = "describe_key_pairs"
        self.describe_kwarg_name = "KeyPairIds"
        self.describe_kwarg_is_list = True
        self.object_path = ".KeyPairs[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )
